# SPDX-License-Identifier: MIT
# Copyright © 2023 Apple Inc.

# Standard
from typing import Optional
import copy

# Third Party
from mlx.utils import tree_flatten
import mlx.core as mx
import mlx.nn as nn
import numpy as np
import torch

# Local
from . import utils


def quantize_model(weights, config, q_group_size, q_bits):
    quantized_config = copy.deepcopy(config)

    # Get model classes
    model_class, model_args_class = utils._get_classes(config=config)

    # Load the model:
    model = model_class(model_args_class.from_dict(config))
    model.load_weights(list(weights.items()))

    # Quantize the model:
    nn.QuantizedLinear.quantize_module(
        model,
        q_group_size,
        q_bits,
        linear_class_predicate=lambda m: isinstance(m, nn.Linear)
        and m.weight.shape[0] != 8,
    )

    # Update the config:
    quantized_config["quantization"] = {
        "group_size": q_group_size,
        "bits": q_bits,
    }
    quantized_weights = dict(tree_flatten(model.parameters()))

    return quantized_weights, quantized_config


def convert_between_mlx_and_pytorch(
    hf_path: str,
    mlx_path: str = "mlx_model",
    quantize: bool = False,
    q_group_size: int = 64,
    q_bits: int = 4,
    dtype: str = "float16",
    upload_name: Optional[str] = None,
    to_pt: bool = False,
    local: bool = False,
):
    """Convert Hugging Face model to MLX format"""
    if dtype not in ("float16", "bfloat16", "float32"):
        raise  # TODO something

    print("[INFO] Loading")
    weights, config, tokenizer = utils.fetch_from_hub(hf_path, local)

    if to_pt:
        dtype = np.float16 if quantize else getattr(np, dtype)
        print(f"{dtype=}")
        weights = {
            k: torch.from_numpy(np.array(v, copy=False, dtype=dtype))
            for k, v in weights.items()
        }
    else:
        dtype = mx.float16 if quantize else getattr(mx, dtype)
        print(f"{dtype=}")
        weights = {k: v.astype(dtype) for k, v in weights.items()}
    if quantize:
        print("[INFO] Quantizing")
        weights, config = quantize_model(weights, config, q_group_size, q_bits)

    utils.save_model(mlx_path, weights, tokenizer, config)
    if upload_name is not None:
        utils.upload_to_hub(mlx_path, upload_name, hf_path)



# SPDX-License-Identifier: Apache-2.0

# Standard
from math import floor
from os import listdir
from os.path import dirname, exists, join
import pathlib
import shutil
import typing

# Third Party
from git import GitError, Repo
import click

# First Party
from instructlab import clickext, utils
from instructlab.configuration import (
    DEFAULTS,
    Config,
    _train,
    ensure_storage_directories_exist,
    get_default_config,
    get_profile_mappings,
    read_config,
    read_train_profile,
    recreate_train_profiles,
    write_config,
)
from instructlab.utils import convert_bytes_to_proper_mag


@click.command()
@click.option(
    "--interactive/--non-interactive",
    default=True,
    show_default=True,
    help="Initialize the environment assuming defaults.",
)
@click.option(
    "--model-path",
    type=click.Path(),
    default=lambda: DEFAULTS.DEFAULT_CHAT_MODEL,
    show_default="The instructlab data files location per the user's system.",
    help="Path to the model used during generation.",
)
@click.option(
    "--taxonomy-base",
    default=DEFAULTS.TAXONOMY_BASE,
    show_default=True,
    help="Base git-ref to use when listing/generating new taxonomy.",
)
@click.option(
    "--taxonomy-path",
    type=click.Path(),
    default=lambda: DEFAULTS.TAXONOMY_DIR,
    show_default="The instructlab data files location per the user's system.",
    help="Path to where the taxonomy should be cloned.",
)
@click.option(
    "--repository",
    default=DEFAULTS.TAXONOMY_REPO,
    show_default=True,
    help="Taxonomy repository location.",
)
@click.option(
    "--min-taxonomy",
    is_flag=True,
    help="Shallow clone the taxonomy repository with minimum size. "
    "Please do not use this option if you are planning to contribute back "
    "using the same taxonomy repository. ",
)
@click.option(
    "--train-profile",
    type=click.Path(),
    default=None,
    help="Overwrite the default training values in the generated config.yaml by passing in an existing training-specific yaml.",
)
@click.option(
    "--config",
    "config_file",
    type=click.Path(),
    default=None,
    envvar=DEFAULTS.ILAB_GLOBAL_CONFIG,
    show_default=True,
    help="Path to a configuration file.",
)
@clickext.display_params
@click.pass_context
def init(
    ctx,
    interactive,
    model_path: str,
    taxonomy_path,
    taxonomy_base,
    repository,
    min_taxonomy,
    train_profile,
    config_file,
):
    """Initializes environment for InstructLab"""
    fresh_install = ensure_storage_directories_exist()
    param_source = ctx.get_parameter_source("config_file")
    overwrite_profile = False
    if interactive:
        overwrite_profile = check_if_configs_exist(fresh_install)
    model_path, taxonomy_path, cfg = get_params(
        param_source,
        interactive,
        repository,
        min_taxonomy,
        model_path,
        taxonomy_path,
        config_file,
    )
    if overwrite_profile:
        click.echo(
            f"\nGenerating config file and profiles:\n    {DEFAULTS.CONFIG_FILE}\n    {DEFAULTS.TRAIN_PROFILE_DIR}\n"
        )
        recreate_train_profiles(overwrite=True)
    else:
        click.echo(f"\nGenerating config file:\n    {DEFAULTS.CONFIG_FILE}\n")
    if train_profile is not None:
        cfg.train = read_train_profile(train_profile)
    elif interactive:
        (
            chosen_profile,
            chosen_vram,
            detected_train_profile,
            edited_cfg,
        ) = hw_auto_detect()
        yn = None
        if chosen_vram is not None and chosen_profile is not None:
            click.echo(
                f"We chose {chosen_profile} as your designated training profile. This is for systems with {chosen_vram} GB of vRAM."
            )
            if edited_cfg:
                click.echo(
                    "This profile is the best approximation for your system based off of the amount of vRAM. We modified it to match the number of GPUs you have."
                )
            yn = click.confirm("Is this profile correct?", default=True)
        if yn is None or not yn:
            train_profile_filenames = sorted(listdir(DEFAULTS.TRAIN_PROFILE_DIR))
            # generate human-friendly text out of train profile filenames
            profile_options = []
            for filename in train_profile_filenames:
                # one gpu type
                split_filename = filename.split("_")
                if len(split_filename) == 2:
                    gputype1 = split_filename[0]
                    num_gpu = split_filename[1].split(".")[0]
                    profile_options.append(f"Nvidia {gputype1} {num_gpu} ({filename})")
                # two gpu types
                elif len(split_filename) == 3:
                    gputype1 = split_filename[0]
                    gputype2 = split_filename[1]
                    num_gpu = split_filename[2].split(".")[0]
                    profile_options.append(
                        f"Nvidia {gputype1}/{gputype2} {num_gpu} ({filename})"
                    )
            click.echo(
                "Please choose a train profile to use.\nTrain profiles assist with the complexity of configuring InstructLab training for specific GPU hardware.\nYou can still take advantage of hardware acceleration for training even if your hardware is not listed."
            )
            click.echo("[0] No profile (CPU, Apple Metal, AMD ROCm)")
            for i, value in enumerate(profile_options):
                click.echo(f"[{i+1}] {value}")
            train_profile_selection = click.prompt(
                "Enter the number of your choice [hit enter for hardware defaults]",
                type=int,
                default=0,
            )
            if 1 <= train_profile_selection <= len(profile_options):
                click.echo(
                    f"You selected: {profile_options[train_profile_selection - 1]}"
                )
                cfg.train = read_train_profile(
                    join(
                        DEFAULTS.TRAIN_PROFILE_DIR,
                        train_profile_filenames[train_profile_selection - 1],
                    )
                )
            elif train_profile_selection == 0:
                click.echo(
                    "No profile selected - any hardware acceleration for training must be configured manually."
                )
            else:
                click.secho(
                    "Invalid selection. Please select a valid train profile option.",
                    fg="red",
                )
                raise click.exceptions.Exit(1)
        else:
            cfg.train = detected_train_profile

    # we should not override all paths with the serve model if special ENV vars exist
    if param_source != click.core.ParameterSource.ENVIRONMENT:
        cfg.chat.model = model_path
        cfg.serve.model_path = model_path
        cfg.evaluate.model = model_path
        cfg.generate.taxonomy_base = taxonomy_base
        cfg.generate.taxonomy_path = taxonomy_path
        cfg.evaluate.mt_bench_branch.taxonomy_path = taxonomy_path
    write_config(cfg)

    ready_text = "  You're ready to start using `ilab`. Enjoy!"
    separator = get_separator(ready_text)
    click.secho(
        f"\n{separator}\n    Initialization completed successfully!\n{ready_text}\n{separator}",
        fg="green",
    )


def hw_auto_detect() -> tuple[str | None, int | None, _train, bool]:
    # Third Party
    import torch

    gpus = 0
    total_vram = 0
    train = _train()
    chosen_vram = None
    chosen_profile_name = None
    gpu_name = ""
    edited_cfg = False
    # try nvidia
    if torch.cuda.is_available() and torch.version.hip is None:
        click.echo("Detecting hardware...")
        gpus = torch.cuda.device_count()
        for i in range(gpus):
            properties = torch.cuda.get_device_properties(i)
            gpu_name = properties.name
            total_vram += properties.total_memory  # memory in B

    vram = int(floor(convert_bytes_to_proper_mag(total_vram)[0]))
    if vram == 0:
        return None, None, train, False
    matched, chosen_vram, chosen_profile_name, train = lookup_card(
        gpu_name=gpu_name, gpu_count=gpus, vram=vram
    )
    # if we used the vRAM to determine, we need to tell the user that
    if not matched:
        # edit nproc per node in the case that we didnt match an exact profile, so we use all GPUs on our system
        train.nproc_per_node = gpus
        edited_cfg = True
    return (chosen_profile_name, chosen_vram, train, edited_cfg)


# lookup_card checks if the user's GPU exists in our supported profile map. If not, we choose one based off of their vRAM
def lookup_card(
    gpu_name, gpu_count, vram
) -> tuple[bool, int | None, str | None, _train]:
    profiles = get_profile_mappings()
    card_entry = profiles.get(gpu_name)
    chosen_profile_name = None
    if card_entry is not None:
        chosen_profile_name = f"Nvidia {gpu_count}x {gpu_name}"
        # it is not guaranteed we match, even if we find an entry
        # what if the vram isn't found or what if the gpu count doesn't match
        matched, train = match_profile_based_on_gpu_count(
            card_entry=card_entry, gpu_count=gpu_count, vram=vram
        )
        if matched:
            return True, vram, chosen_profile_name, train
    vram, chosen_profile_name, train = match_profile_based_on_vram(vram=vram)
    # return False with the results of the vRAM matching to indicate we didn't find an exact match for this GPU configuration in our supported profiles
    return False, vram, chosen_profile_name, train


# match_profile_based_on_vram chooses a training profile based off the amount of vram in each profile
# this profile will be modified to match the number of GPUs on the user's system
def match_profile_based_on_vram(vram) -> tuple[int | None, str | None, _train]:
    chosen_profile_name = None
    chosen_vram = None
    train = _train()
    profiles = get_profile_mappings()
    for group_gpu_name, list_of_count_vram_and_config in profiles.items():
        for gpu_count_configs in list_of_count_vram_and_config:
            if (
                not isinstance(gpu_count_configs["gpu_count"], int)
                or not isinstance(gpu_count_configs["vram_and_config"], dict)
                or not isinstance(gpu_count_configs["vram_and_config"]["vram"], int)
                or not isinstance(
                    gpu_count_configs["vram_and_config"]["config"], _train
                )
            ):
                click.secho("Unable to retrieve train profiles", fg="red")
                raise click.exceptions.Exit(1)
            if (
                chosen_vram is None
                or chosen_vram > gpu_count_configs["vram_and_config"]["vram"]
            ) and vram < gpu_count_configs["vram_and_config"]["vram"]:
                # set our return values
                gpus = gpu_count_configs["gpu_count"]
                chosen_profile_name = f"Nvidia {gpus}x {group_gpu_name}"
                chosen_vram = gpu_count_configs["vram_and_config"]["vram"]
                train = gpu_count_configs["vram_and_config"]["config"]
    return chosen_vram, chosen_profile_name, train


# match_profile_based_on_gpu_count looks for a training profile which directly matches the gpu name and count
# this training profile will be used as is
def match_profile_based_on_gpu_count(
    card_entry, gpu_count, vram
) -> tuple[bool, _train]:
    train = _train()
    matched = False
    for gpu_count_configs in card_entry:
        if not isinstance(gpu_count_configs["gpu_count"], int) or not isinstance(
            gpu_count_configs["vram_and_config"], dict
        ):
            click.secho("Unable to retrieve train profiles", fg="red")
            raise click.exceptions.Exit(1)
        # if GPU name and number matches, this is a direct match
        if (
            gpu_count == gpu_count_configs["gpu_count"]
            and vram == gpu_count_configs["vram_and_config"]["vram"]
        ):
            train = gpu_count_configs["vram_and_config"]["config"]
            matched = True
            break
    return matched, train


def check_if_configs_exist(fresh_install) -> bool:
    if exists(DEFAULTS.CONFIG_FILE):
        click.echo(f"Existing config file was found in:\n    {DEFAULTS.CONFIG_FILE}")
        overwrite = click.confirm("Do you still want to continue?")
        if not overwrite:
            raise click.exceptions.Exit(0)
    if exists(DEFAULTS.TRAIN_PROFILE_DIR) and not fresh_install:
        click.echo(
            f"\nExisting training profiles were found in:\n    {DEFAULTS.TRAIN_PROFILE_DIR}"
        )
        return click.confirm(
            "Do you want to restore these profiles to the default values?"
        )
    # default behavior should be do NOT overwrite files that could have just been created
    return False


def get_params_from_env(
    obj: typing.Optional[typing.Any],
) -> typing.Tuple[str, str, Config]:
    if obj is None or not hasattr(obj, "config"):
        raise ValueError("obj must not be None and must have a 'config' attribute")
    return (
        obj.config.generate.taxonomy_path,
        obj.config.generate.taxonomy_base,
        obj.config,
    )


def get_params(
    param_source: click.core.ParameterSource,
    interactive: bool,
    repository: str,
    min_taxonomy: bool,
    model_path: str,
    taxonomy_path: pathlib.Path,
    config: pathlib.Path | None,
) -> typing.Tuple[str, pathlib.Path, Config]:
    cfg = get_default_config()
    if config is not None:
        cfg = read_config(config)
    clone_taxonomy_repo = True
    if interactive:
        guide_text = "  This guide will help you to setup your environment"
        separator = get_separator(guide_text)
        click.echo(
            f"\n{separator}\n         Welcome to the InstructLab CLI\n{guide_text}\n{separator}\n"
        )
        click.echo(
            "Please provide the following values to initiate the "
            "environment [press Enter for defaults]:"
        )
        if param_source != click.core.ParameterSource.ENVIRONMENT:
            taxonomy_path = utils.expand_path(
                click.prompt("Path to taxonomy repo", default=taxonomy_path)
            )

    try:
        taxonomy_contents = listdir(taxonomy_path)
    except FileNotFoundError:
        taxonomy_contents = []
    if taxonomy_contents:
        clone_taxonomy_repo = False
    elif interactive and param_source != click.core.ParameterSource.ENVIRONMENT:
        click.echo(f"`{taxonomy_path}` seems to not exist or is empty.")
        clone_taxonomy_repo = click.confirm(
            f"Should I clone {repository} for you?",
            default=True,
        )

    # clone taxonomy repo if it needs to be cloned
    if clone_taxonomy_repo:
        click.echo(f"Cloning {repository}...")
        clone_depth = False if not min_taxonomy else 1
        try:
            Repo.clone_from(
                repository,
                taxonomy_path,
                branch="main",
                recurse_submodules=True,
                depth=clone_depth,
            )
        except GitError as exc:
            click.secho(f"Failed to clone taxonomy repo: {exc}", fg="red")
            click.secho(f"Please make sure to manually run `git clone {repository}`")
            raise click.exceptions.Exit(1)

    # check if models dir exists, and if so ask for which model to use
    models_dir = dirname(model_path)
    if (
        interactive
        and exists(models_dir)
        and param_source != click.core.ParameterSource.ENVIRONMENT
    ):
        model_path = utils.expand_path(
            click.prompt("Path to your model", default=model_path)
        )

    return model_path, taxonomy_path, cfg


def get_separator(text: str) -> str:
    text_length = len(text)
    try:
        terminal_width = shutil.get_terminal_size().columns
    except Exception:  # pylint: disable=broad-exception-caught
        # Exception can occur in non-interactive scenarios where no terminal is associated with the process
        terminal_width = text_length
    separator_length = min(text_length, terminal_width)
    return "-" * separator_length



# SPDX-License-Identifier: Apache-2.0

# Standard
import logging
import pathlib

# Third Party
from click.testing import CliRunner
import yaml

# First Party
from instructlab import configuration, lab


def test_ilab_config_show(cli_runner: CliRunner) -> None:
    result = cli_runner.invoke(lab.ilab, ["--config", "DEFAULT", "config", "show"])
    assert result.exit_code == 0, result.stdout

    parsed = yaml.safe_load(result.stdout_bytes)
    assert parsed

    assert configuration.Config(**parsed)


def test_ilab_config_init_with_env_var_config(
    cli_runner: CliRunner, tmp_path: pathlib.Path
) -> None:
    # Common setup code
    cfg = configuration.get_default_config()
    assert cfg.general.log_level == logging.getLevelName(logging.INFO)
    cfg.general.log_level = logging.getLevelName(logging.DEBUG)
    cfg_file = tmp_path / "config-gold.yaml"
    with cfg_file.open("w") as f:
        yaml.dump(cfg.model_dump(), f)

    # Invoke the CLI command
    command = ["config", "init", "--non-interactive"]
    result = cli_runner.invoke(
        lab.ilab, command, env={"ILAB_GLOBAL_CONFIG": cfg_file.as_posix()}
    )
    assert result.exit_code == 0, result.stdout

    # Load and check the generated config file
    config_path = pathlib.Path(configuration.DEFAULTS.CONFIG_FILE)
    assert config_path.exists()
    with config_path.open(encoding="utf-8") as f:
        parsed = yaml.safe_load(f)
    assert parsed
    assert configuration.Config(**parsed).general.log_level == logging.getLevelName(
        logging.DEBUG
    )


def test_ilab_config_init_with_model_path(cli_runner: CliRunner) -> None:
    # Common setup code
    model_path = "path/to/model"
    command = ["config", "init", "--model-path", model_path]

    # Invoke the CLI command
    result = cli_runner.invoke(lab.ilab, command)
    assert result.exit_code == 0, result.stdout

    # Load and check the generated config file
    config_path = pathlib.Path(configuration.DEFAULTS.CONFIG_FILE)
    assert config_path.exists()
    with config_path.open(encoding="utf-8") as f:
        parsed = yaml.safe_load(f)
    assert parsed
    # the generate config should NOT use the same model as the chat/serve model
    assert configuration.Config(**parsed).generate.model != "path/to/model"
    assert configuration.Config(**parsed).generate.teacher.model_path != "path/to/model"
    assert configuration.Config(**parsed).serve.model_path == "path/to/model"



# SPDX-License-Identifier: MIT

# Standard
import json

# Local
from ...utils import get_sysprompt


def format_text(obj):
    return f"""\
<|system|>
{obj['system']}
<|user|>
{obj['user']}
<|assistant|>
{obj['assistant']}<|endoftext|>\
"""


def make_data(data_dir: str, is_shiv: bool = False):
    if not is_shiv:
        # This branch uses data from `lab generate`
        # train_gen.jsonl and test_gen.jsonl are the two files produced by `lab generate`
        for filename in [f"{data_dir}/train_gen.jsonl", f"{data_dir}/test_gen.jsonl"]:
            # Load the JSON Lines file
            with open(filename, "r", encoding="utf-8") as f:
                data = [json.loads(line) for line in f]

            # Add the "text" field with value "x" to each object
            data_new = []
            for obj in data:
                obj_new = {
                    "system": get_sysprompt(),
                    "user": obj["user"],
                    "assistant": obj["assistant"],
                }
                data_new.append(obj_new | {"text": format_text(obj_new)})

            # Save the modified objects back to the JSON Lines file
            if "train_gen" in filename:
                n = len(data_new) * 8 // 10
                with open(f"{data_dir}/train.jsonl", "w", encoding="utf-8") as f:
                    for obj in data_new[:n]:
                        f.write(json.dumps(obj) + "\n")
                with open(f"{data_dir}/valid.jsonl", "w", encoding="utf-8") as f:
                    for obj in data_new[n:]:
                        f.write(json.dumps(obj) + "\n")
            if "test_gen" in filename:
                with open(f"{data_dir}/test.jsonl", "w", encoding="utf-8") as f:
                    for obj in data_new:
                        f.write(json.dumps(obj) + "\n")
    else:
        # This branch is to use Shiv generated data
        # You can ignore for now
        filename = f"{data_dir}/raw.jsonl"

        # Load the JSON Lines file
        with open(filename, "r", encoding="utf-8") as f:
            data = [json.loads(line) for line in f]

        # Add the "text" field with value "x" to each object
        data_new = []
        for obj in data:
            obj_new = {
                "system": get_sysprompt(),
                "user": obj["inputs"],
                "assistant": obj["targets"],
            }
            data_new.append(obj_new | {"text": format_text(obj_new)})

        # Save the modified objects back to the JSON Lines file
        n = len(data_new) // 10 * 7
        m = len(data_new) // 10 * 2 + n
        with open(f"{data_dir}/train.jsonl", "w", encoding="utf-8") as f:
            for obj in data_new[:n]:
                f.write(json.dumps(obj) + "\n")
        with open(f"{data_dir}/valid.jsonl", "w", encoding="utf-8") as f:
            for obj in data_new[n:m]:
                f.write(json.dumps(obj) + "\n")
        with open(f"{data_dir}/test.jsonl", "w", encoding="utf-8") as f:
            for obj in data_new[:m]:
                f.write(json.dumps(obj) + "\n")



# SPDX-License-Identifier: Apache-2.0
#
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "InstructLab"
copyright = "2024, InstructLab Authors"
author = "InstructLab Authors"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_click",
]

templates_path = ["templates"]
exclude_patterns = [
    "build",
    "Thumbs.db",
    ".DS_Store",
    "cli_reference.md",  # ignored in favor of auto-generated docs
]

intersphinx_mapping = {
    # "python": ("https://docs.python.org/", None),
    # "torch": ("https://pytorch.org/docs/stable/", None),
    # "click": ("https://click.palletsprojects.com/en/latest/", None),
    # "pydantic": ("https://docs.pydantic.dev/latest/", None),
    # "transformers": ("https://huggingface.co/docs/transformers/main/en", None),
}

autodoc_mock_imports = [
    "fire",
    "mlx",
]

nitpick_ignore = [
    ("py:class", "git.repo.base.Repo"),
    ("py:class", "uvicorn.config.Config"),
    ("py:class", "uvicorn.server.Server"),
    ("py:class", "transformers.generation.stopping_criteria.StoppingCriteria"),
    ("py:class", "mlx.core.array"),
    ("py:class", "mlx.nn.LayerNorm"),
    ("py:class", "mlx.nn.Linear"),
    ("py:class", "mlx.nn.Module"),
    ("py:class", "nn.Module"),
    ("py:class", "mx.array"),
    # stdlib
    ("py:class", "FrameType"),
    # pydantic auto-generated doc strings
    ("py:class", "ComputedFieldInfo"),
    ("py:class", "ConfigDict"),
    ("py:class", "FieldInfo"),
]

myst_heading_anchors = 3

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["static"]



# SPDX-License-Identifier: MIT

# Standard
import math

# Third Party
import mlx.core as mx
import mlx.nn as nn


class LoRALinear(nn.Module):
    @staticmethod
    def from_linear(linear: nn.Linear, rank: int = 8):
        # TODO remove when input_dims and output_dims are attributes
        # on linear and quantized linear
        output_dims, input_dims = linear.weight.shape
        if isinstance(linear, nn.QuantizedLinear):
            input_dims *= 32 // linear.bits
        lora_lin = LoRALinear(input_dims, output_dims, rank)
        lora_lin.linear = linear
        return lora_lin

    def to_linear(self, de_quantize: bool = False):
        linear = self.linear
        bias = "bias" in linear
        weight = linear.weight
        is_quantized = isinstance(linear, nn.QuantizedLinear)

        # Use the same type as the linear weight if not quantized
        dtype = weight.dtype

        if is_quantized:
            dtype = mx.float16
            weight = mx.dequantize(
                weight,
                linear.scales,
                linear.biases,
                linear.group_size,
                linear.bits,
            )
        output_dims, input_dims = weight.shape
        fused_linear = nn.Linear(input_dims, output_dims, bias=bias)

        lora_b = (self.scale * self.lora_b.T).astype(dtype)
        lora_a = self.lora_a.T.astype(dtype)
        fused_linear.weight = weight + lora_b @ lora_a
        if bias:
            fused_linear.bias = linear.bias

        if is_quantized and not de_quantize:
            fused_linear = nn.QuantizedLinear.from_linear(
                fused_linear,
                linear.group_size,
                linear.bits,
            )

        return fused_linear

    def __init__(
        self,
        input_dims: int,
        output_dims: int,
        lora_rank: int = 8,
        bias: bool = False,
        scale: float = 20.0,
    ):
        super().__init__()

        # Regular linear layer weights
        self.linear = nn.Linear(input_dims, output_dims, bias=bias)

        # Scale for low-rank update
        self.scale = scale

        # Low rank lora weights
        scale = 1 / math.sqrt(input_dims)
        self.lora_a = mx.random.uniform(
            low=-scale,
            high=scale,
            shape=(input_dims, lora_rank),
        )
        self.lora_b = mx.zeros(shape=(lora_rank, output_dims))

    def __call__(self, x):
        dtype = self.linear.weight.dtype
        if isinstance(self.linear, nn.QuantizedLinear):
            dtype = self.linear.scales.dtype
        y = self.linear(x.astype(dtype))
        z = (x @ self.lora_a) @ self.lora_b
        return y + self.scale * z



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
from typing import List
import logging

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab.configuration import DEFAULTS
from instructlab.utils import list_models, print_table

logger = logging.getLogger(__name__)


@click.command(name="list")
@clickext.display_params
@click.option(
    "--model-dirs",
    help="Base directories where models are stored.",
    multiple=True,
    default=[DEFAULTS.MODELS_DIR],
    show_default=True,
)
@click.option(
    "--list-checkpoints",
    help="Also list checkpoints [in addition to existing models].",
    is_flag=True,
)
def model_list(model_dirs: List[str], list_checkpoints: bool):
    """Lists models"""
    # click converts lists into tuples when using multiple
    model_path_dirs = [Path(m) for m in model_dirs]

    data = list_models(model_path_dirs, list_checkpoints)
    data_as_lists = [list(item) for item in data]  # this is to satisfy mypy
    print_table(["Model Name", "Last Modified", "Size"], data_as_lists)



# SPDX-License-Identifier: MIT
# Copyright © 2023 Apple Inc.

# Standard
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

# Third Party
from huggingface_hub import snapshot_download
import mlx.core as mx
import mlx.nn as nn

# Local
from .utils import save_model


@dataclass
class ModelArgs:
    hidden_size: int
    num_hidden_layers: int
    intermediate_size: int
    num_attention_heads: int
    rms_norm_eps: float
    vocab_size: int
    num_key_value_heads: int = None
    rope_theta: float = 10000
    rope_traditional: bool = False
    model_type: str = None
    rope_scaling: Optional[Dict[str, Union[float, str]]] = None

    def __post_init__(self):
        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads

        if self.rope_scaling:
            required_keys = {"factor", "type"}
            if not all(key in self.rope_scaling for key in required_keys):
                raise ValueError(f"rope_scaling must contain keys {required_keys}")

            if self.rope_scaling["type"] != "linear":
                raise ValueError("rope_scaling 'type' currently only supports 'linear'")


class RMSNorm(nn.Module):
    def __init__(self, dims: int, eps: float = 1e-5):
        super().__init__()
        self.weight = mx.ones((dims,))
        self.eps = eps

    def _norm(self, x):
        return x * mx.rsqrt(x.square().mean(-1, keepdims=True) + self.eps)

    def __call__(self, x):
        output = self._norm(x.astype(mx.float32)).astype(x.dtype)
        return self.weight * output


class Attention(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()

        dim = args.hidden_size
        self.n_heads = n_heads = args.num_attention_heads
        self.n_kv_heads = n_kv_heads = args.num_key_value_heads

        self.repeats = n_heads // n_kv_heads

        head_dim = args.hidden_size // n_heads
        self.scale = head_dim**-0.5

        self.q_proj = nn.Linear(dim, n_heads * head_dim, bias=False)
        self.k_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=False)
        self.v_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=False)
        self.o_proj = nn.Linear(n_heads * head_dim, dim, bias=False)
        rope_scale = (
            1 / args.rope_scaling["factor"]
            if args.rope_scaling is not None and args.rope_scaling["type"] == "linear"
            else 1
        )
        self.rope = nn.RoPE(
            head_dim,
            traditional=args.rope_traditional,
            base=args.rope_theta,
            scale=rope_scale,
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Tuple[mx.array, mx.array]] = None,
    ) -> mx.array:
        B, L, D = x.shape

        queries, keys, values = self.q_proj(x), self.k_proj(x), self.v_proj(x)

        # Prepare the queries, keys and values for the attention computation
        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if self.repeats > 1:
            keys = mx.repeat(keys, self.repeats, axis=1)
            values = mx.repeat(values, self.repeats, axis=1)

        if cache is not None:
            key_cache, value_cache = cache
            queries = self.rope(queries, offset=key_cache.shape[2])
            keys = self.rope(keys, offset=key_cache.shape[2])
            keys = mx.concatenate([key_cache, keys], axis=2)
            values = mx.concatenate([value_cache, values], axis=2)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        scores = (queries * self.scale) @ keys.transpose(0, 1, 3, 2)
        if mask is not None:
            scores += mask
        scores = mx.softmax(scores.astype(mx.float32), axis=-1).astype(scores.dtype)
        output = (scores @ values).transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output), (keys, values)


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim):
        super().__init__()
        self.gate_proj = nn.Linear(dim, hidden_dim, bias=False)
        self.down_proj = nn.Linear(hidden_dim, dim, bias=False)
        self.up_proj = nn.Linear(dim, hidden_dim, bias=False)

    def __call__(self, x) -> mx.array:
        return self.down_proj(nn.silu(self.gate_proj(x)) * self.up_proj(x))


class TransformerBlock(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.num_attention_heads = args.num_attention_heads
        self.hidden_size = args.hidden_size
        self.self_attn = Attention(args)
        self.mlp = MLP(args.hidden_size, args.intermediate_size)
        self.input_layernorm = RMSNorm(args.hidden_size, eps=args.rms_norm_eps)
        self.post_attention_layernorm = RMSNorm(args.hidden_size, eps=args.rms_norm_eps)
        self.args = args

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Tuple[mx.array, mx.array]] = None,
    ) -> mx.array:
        r, cache = self.self_attn(self.input_layernorm(x), mask, cache)
        h = x + r
        r = self.mlp(self.post_attention_layernorm(h))
        out = h + r
        return out, cache


class LlamaModel(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.args = args
        self.vocab_size = args.vocab_size
        self.num_hidden_layers = args.num_hidden_layers
        assert self.vocab_size > 0
        self.embed_tokens = nn.Embedding(args.vocab_size, args.hidden_size)
        self.layers = [
            TransformerBlock(args=args) for _ in range(args.num_hidden_layers)
        ]
        self.norm = RMSNorm(args.hidden_size, eps=args.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        cache=None,
    ):
        h = self.embed_tokens(inputs)

        mask = None
        if h.shape[1] > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(h.shape[1])
            mask = mask.astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h, cache[e] = layer(h, mask, cache[e])

        return self.norm(h), cache


class Model(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.model = LlamaModel(args)
        self.lm_head = nn.Linear(args.hidden_size, args.vocab_size, bias=False)

    def __call__(
        self,
        inputs: mx.array,
        cache=None,
    ):
        out, cache = self.model(inputs, cache)
        return self.lm_head(out), cache


def get_config(metadata: dict):
    output = {
        "hidden_size": metadata["llama.embedding_length"],
        "num_hidden_layers": metadata["llama.block_count"],
        "num_attention_heads": metadata["llama.attention.head_count"],
        "intermediate_size": metadata["llama.feed_forward_length"],
        "num_key_value_heads": metadata["llama.attention.head_count_kv"],
        "rms_norm_eps": metadata["llama.attention.layer_norm_rms_epsilon"],
        "vocab_size": len(metadata["tokenizer.ggml.tokens"]),
        "rope_theta": metadata["llama.rope.freq_base"],
        "rope_traditional": True,
    }
    output = {k: v.item() if isinstance(v, mx.array) else v for k, v in output.items()}
    return output


def translate_weight_names(name):
    name = name.replace("blk.", "model.layers.")
    name = name.replace("ffn_gate", "mlp.gate_proj")
    name = name.replace("ffn_down", "mlp.down_proj")
    name = name.replace("ffn_up", "mlp.up_proj")
    name = name.replace("attn_q", "self_attn.q_proj")
    name = name.replace("attn_k", "self_attn.k_proj")
    name = name.replace("attn_v", "self_attn.v_proj")
    name = name.replace("attn_output", "self_attn.o_proj")
    name = name.replace("attn_norm", "input_layernorm")
    name = name.replace("ffn_norm", "post_attention_layernorm")
    name = name.replace("token_embd", "model.embed_tokens")
    name = name.replace("output_norm", "model.norm")
    name = name.replace("output", "lm_head")
    return name


def load(gguf: str, repo: Optional[str] = None, mlx_path: Optional[str] = None):
    """Inference script"""
    # If the gguf exists, try to load model from it.
    # Otherwise try to download and cache from the HF repo
    if not Path(gguf).exists():
        if repo is None:
            raise ValueError(
                f"Could not find file {gguf}, and no Hugging Face"
                " repo provided for download."
            )
        model_path = snapshot_download(
            repo_id=repo,
            allow_patterns=[gguf],
        )
        if not (Path(model_path) / gguf).exists():
            raise ValueError(f"File {gguf} not in repo {repo}.")
        gguf = str(Path(model_path) / gguf)

    print(f"[INFO] Loading model from {gguf}")
    weights, metadata = mx.load(gguf, return_metadata=True)
    gguf_ft = metadata["general.file_type"]
    if gguf_ft == 0 or gguf_ft == 1:
        # ALL_F32 or MOSTLY_F16
        quantization = None
        pass
    elif gguf_ft == 2 or gguf_ft == 3:
        # MOSTLY_Q4_0 or MOSTLY_Q4_1
        quantization = {"group_size": 32, "bits": 4}
    elif gguf_ft == 7:
        # MOSTLY_Q8_0 = 7
        quantization = {"group_size": 32, "bits": 8}
    else:
        quantization = None
        print("[WARNING] Using unsupported GGUF quantization. Casting to float16.")

    weights = {translate_weight_names(k): v for k, v in weights.items()}
    config = get_config(metadata)
    model = Model(ModelArgs(**config))
    if quantization is not None:
        # quantized the LM head?
        qm = model if "lm_head.scales" in weights else model.model
        nn.QuantizedLinear.quantize_module(
            qm,
            **quantization,
        )

    model.load_weights(list(weights.items()))
    save_model(mlx_path, weights)


def generate(prompt: mx.array, model: Model, temp: float = 0.0):
    def sample(logits):
        if temp == 0:
            return mx.argmax(logits, axis=-1)
        else:
            return mx.random.categorical(logits * (1 / temp))

    y = prompt
    cache = None
    while True:
        logits, cache = model(y[None], cache=cache)
        logits = logits[:, -1, :]
        y = sample(logits)
        yield y



# SPDX-License-Identifier: MIT
# Copyright © 2023 Apple Inc.

# Standard
from pathlib import Path
from typing import Optional
import json
import math
import time

# Third Party
from mlx.utils import tree_flatten
import mlx.core as mx
import mlx.nn as nn
import mlx.optimizers as optim
import numpy as np

# Local
from . import utils as lora_utils
from .models.lora import LoRALinear


class Dataset:
    """
    Light-weight wrapper to hold lines from a jsonl file
    """

    def __init__(self, path: Path, key: str = "text"):
        if not path.exists():
            self._data = None
        else:
            with open(path, "r") as fid:
                self._data = [json.loads(l) for l in fid]
        self._key = key

    def __getitem__(self, idx: int):
        return self._data[idx][self._key]

    def __len__(self):
        return len(self._data)


def load(data, train, test):
    def load_and_check(name):
        dataset_path = Path(data) / f"{name}.jsonl"
        try:
            return Dataset(dataset_path)
        except Exception as e:
            print(f"Unable to build dataset {dataset_path} ({e})")
            raise

    names = ("train", "valid", "test")
    train_dataset_path, valid_dataset_path, test_dataset_path = (
        load_and_check(n) for n in names
    )

    if train and len(train_dataset_path) == 0:
        raise ValueError(
            "Training set not found or empty. Must provide training set for fine-tuning."
        )
    if train and len(valid_dataset_path) == 0:
        raise ValueError(
            "Validation set not found or empty. Must provide validation set for fine-tuning."
        )
    if test and len(test_dataset_path) == 0:
        raise ValueError(
            "Test set not found or empty. Must provide test set for evaluation."
        )
    return train_dataset_path, valid_dataset_path, test_dataset_path


def loss(model, inputs, targets, lengths):
    # Run model on inputs
    logits, _ = model(inputs)
    logits = logits.astype(mx.float32)

    # Mask padding tokens
    length_mask = mx.arange(inputs.shape[1])[None, :] < lengths[:, None]

    # Calculate the loss
    ce = nn.losses.cross_entropy(logits, targets) * length_mask
    ntoks = length_mask.sum()
    ce = ce.sum() / ntoks
    return ce, ntoks


def iterate_batches(dset, tokenizer, batch_size, train=False):
    # Shuffle indices
    while True:
        indices = np.arange(len(dset))
        if train:
            indices = np.random.permutation(indices)

        # Collect batches from dataset
        for i in range(0, len(indices) - batch_size + 1, batch_size):
            # Encode batch
            batch = [tokenizer.encode(dset[indices[i + j]]) for j in range(batch_size)]
            lengths = [len(x) for x in batch]

            # Check if any sequence is longer than 2048 tokens
            if max(lengths) > 2048:
                print(
                    "[WARNING] Some sequences are longer than 2048 tokens. "
                    "Consider pre-splitting your data to save memory."
                )

            # Pad to the max length
            batch_arr = np.zeros((batch_size, max(lengths)), np.int32)

            for j in range(batch_size):
                batch_arr[j, : lengths[j]] = batch[j]
            batch = mx.array(batch_arr)
            yield batch[:, :-1], batch[:, 1:], mx.array(lengths)

        if not train:
            break


def evaluate(model, dataset, loss, tokenizer, batch_size, num_batches):
    all_losses = []
    ntokens = 0
    for it, batch in zip(
        range(num_batches),
        iterate_batches(dataset, tokenizer, batch_size),
    ):
        losses, toks = loss(model, *batch)
        all_losses.append((losses * toks).item())
        ntokens += toks.item()

    return np.sum(all_losses) / ntokens


def train_model(
    model,
    train_set,
    val_set,
    optimizer,
    loss,
    tokenizer,
    iters,
    batch_size,
    steps_per_report,
    steps_per_eval,
    val_batches,
    save_every,
    adapter_file,
):
    # Create value and grad function for loss
    loss_value_and_grad = nn.value_and_grad(model, loss)

    losses = []
    n_tokens = 0

    # Main training loop
    start = time.perf_counter()
    for it, batch in zip(
        range(iters),
        iterate_batches(train_set, tokenizer, batch_size, train=True),
    ):
        # Forward and backward pass
        (lvalue, toks), grad = loss_value_and_grad(model, *batch)

        # Model update
        optimizer.update(model, grad)
        mx.eval(model.parameters(), optimizer.state, lvalue)

        # Record loss
        losses.append(lvalue.item())
        n_tokens += toks.item()

        # Report training loss if needed
        if (it + 1) % steps_per_report == 0:
            train_loss = np.mean(losses)

            stop = time.perf_counter()
            print(
                f"Iter {it+1:03d}: Train loss {train_loss:.3f}, "
                f"It/sec {steps_per_report / (stop - start):.3f}, "
                f"Tokens/sec {float(n_tokens) / (stop - start):.3f}"
            )
            losses = []
            n_tokens = 0
            start = time.perf_counter()

        # Report validation loss if needed
        if it == 0 or (it + 1) % steps_per_eval == 0:
            stop = time.perf_counter()
            val_loss = evaluate(
                model, val_set, loss, tokenizer, batch_size, val_batches
            )
            epoch = (it * batch_size) // len(train_set)
            print(
                f"Epoch {epoch + 1}: "
                f"Iter {it + 1}: "
                f"Val loss {val_loss:.3f}, "
                f"Val took {(time.perf_counter() - stop):.3f}s"
            )

            start = time.perf_counter()

        # Save adapter weights if needed
        if (it + 1) % save_every == 0:
            mx.savez(adapter_file, **dict(tree_flatten(model.trainable_parameters())))
            # Update the split function to based on the last occurence of "." since the file path now includes . in MacOs.
            a, b = adapter_file.rsplit(".", 1)
            fn = f"{a}-{it+1:03d}.{b}"
            mx.savez(fn, **dict(tree_flatten(model.trainable_parameters())))
            print(f"Iter {it + 1}: Saved adapter weights to {fn}.")


def generate(model, prompt, tokenizer, stream, temp, max_tokens):
    if stream:
        print(prompt, end="", flush=True)

    prompt = mx.array(tokenizer.encode(prompt))

    tokens = []
    skip = 0
    for token, n in zip(
        lora_utils.generate(prompt, model, temp),
        range(max_tokens),
    ):
        if token == tokenizer.eos_token_id:
            break

        tokens.append(token.item())
        s = tokenizer.decode(tokens)
        if len(s) - skip > 1:
            if stream:
                print(s[skip:-1], end="", flush=True)
            skip = len(s) - 1
    if stream:
        print(tokenizer.decode(tokens)[skip:], flush=True)
    else:
        print("=" * 10)
        print(tokenizer.decode(tokens).strip())
        print("=" * 10)
    if len(tokens) == 0:
        print("No tokens generated for this prompt")
        return


def load_and_train(
    model: str = "mlx_model",
    max_tokens: int = 100,
    temp: float = 0.8,
    prompt: Optional[str] = None,
    train: bool = False,
    data: str = "data/",
    lora_layers: int = 16,
    batch_size: int = 4,
    iters: int = 1000,
    val_batches: int = 25,
    learning_rate: float = 1e-5,
    steps_per_report: int = 10,
    steps_per_eval: int = 200,
    resume_adapter_file: Optional[str] = None,
    adapter_file: str = "adapters.npz",
    save_every: int = 100,
    test: bool = False,
    stream: bool = False,
    no_adapter: bool = False,
    test_batches: int = 500,
    seed: int = 0,
):
    """LoRA or QLoRA fine tuning."""
    np.random.seed(seed)

    print("Loading pretrained model")
    model, tokenizer, _ = lora_utils.load(model)

    # Freeze all layers other than LORA linears
    model.freeze()
    if not no_adapter:
        for l in model.model.layers[len(model.model.layers) - lora_layers :]:
            l.self_attn.q_proj = LoRALinear.from_linear(l.self_attn.q_proj)
            l.self_attn.v_proj = LoRALinear.from_linear(l.self_attn.v_proj)
            if hasattr(l, "block_sparse_moe"):
                l.block_sparse_moe.gate = LoRALinear.from_linear(
                    l.block_sparse_moe.gate
                )
    else:
        print("LoRA init skipped")

    p = sum(v.size for _, v in tree_flatten(model.parameters())) / 10**6
    print(f"Total parameters {p:.3f}M")
    p = sum(v.size for _, v in tree_flatten(model.trainable_parameters())) / 10**6
    print(f"Trainable parameters {p:.3f}M")

    print("Loading datasets")
    train_set, valid_set, test_set = load(data, train, test)

    # Resume training the given adapters.
    if resume_adapter_file is not None:
        print(f"Loading pretrained adapters from {resume_adapter_file}")
        model.load_weights(resume_adapter_file, strict=False)

    if train:
        print("*********")
        print("")
        print("ᕙ(•̀‸•́‶)ᕗ  Training has started! ᕙ(•̀‸•́‶)ᕗ ")
        print("")
        print("*********")
        opt = optim.Adam(learning_rate=learning_rate)

        # Train model
        train_model(
            model,
            train_set,
            valid_set,
            opt,
            loss,
            tokenizer,
            iters,
            batch_size,
            steps_per_report,
            steps_per_eval,
            val_batches,
            save_every,
            adapter_file,
        )

        # Save adapter weights
        mx.savez(adapter_file, **dict(tree_flatten(model.trainable_parameters())))

    if not no_adapter:
        # Load the LoRA adapter weights which we assume should exist by this point
        if not Path(adapter_file).is_file():
            raise ValueError(
                f"Adapter file {adapter_file} missing. "
                "Use --train to learn and save the adapters.npz."
            )
        model.load_weights(adapter_file, strict=False)
    else:
        print("LoRA loading skipped")

    if test:
        print("Testing")
        model.eval()
        test_loss = evaluate(
            model,
            test_set,
            loss,
            tokenizer,
            batch_size,
            num_batches=test_batches,
        )
        test_ppl = math.exp(test_loss)

        print(f"Test loss {test_loss:.3f}, Test ppl {test_ppl:.3f}.")

    if prompt is not None:
        print("Generating")
        generate(model, prompt, tokenizer, stream, temp, max_tokens)



# SPDX-License-Identifier: Apache-2.0
# pylint: disable=redefined-outer-name

# Standard
from unittest import mock
import os
import pathlib
import sys
import typing

# Third Party
from click.testing import CliRunner
import pytest

# Local
from .taxonomy import MockTaxonomy

TESTS_PATH = pathlib.Path(__file__).parent.absolute()


def pytest_addoption(parser) -> None:
    parser.addoption(
        "--regenerate-testdata",
        action="store_true",
        help="Regenerate test data like dumped config",
    )


@pytest.fixture
def regenerate_testdata(request) -> bool:
    return bool(request.config.getoption("--regenerate-testdata"))


@pytest.fixture
def taxonomy_dir(tmp_path):
    with MockTaxonomy(tmp_path) as taxonomy:
        yield taxonomy


@pytest.fixture(scope="class")
def mock_mlx_package():
    """Add mocked 'mlx' modules to sys.path

    The 'mlx' package is only available on Apple Silicon. This fixture
    patches sys.modules so local imports and mock-patching of 'mlx' attributes
    works.
    """
    mlx_modules = {
        name: mock.MagicMock()
        for name in ["mlx", "mlx.core", "mlx.nn", "mlx.optimizers", "mlx.utils"]
    }
    with mock.patch.dict(sys.modules, mlx_modules):
        yield


@pytest.fixture
def tmp_path_home(tmp_path: pathlib.Path) -> typing.Generator[pathlib.Path, None, None]:
    """Reset $HOME to tmp_path and unset $XDG_*

    Yields `tmp_path` fixture path.
    """
    # First Party
    from instructlab import lab
    from instructlab.configuration import DEFAULTS

    # find the --config-file option
    opt_config_file = [opt for opt in lab.ilab.params if opt.name == "config_file"][0]

    with mock.patch.dict(os.environ):
        # set HOME as home of all XDG directories
        os.environ["HOME"] = str(tmp_path)
        # unset custom config location
        os.environ.pop(DEFAULTS.ILAB_GLOBAL_CONFIG, None)
        # unset all XDG env vars
        for key in list(os.environ):
            if key.startswith("XDG_"):
                os.environ.pop(key)
        # resets the config defaults to use the new $HOME
        DEFAULTS._reset()
        # patch default value of --config-file option
        with mock.patch.object(
            opt_config_file, attribute="default", new=DEFAULTS.CONFIG_FILE
        ):
            yield tmp_path


@pytest.fixture
def cli_runner(
    tmp_path_home: pathlib.Path,
) -> typing.Generator[CliRunner, None, None]:
    """Click CLI runner

    Run click's CliRunner with file system isolation (chdir), $HOME reset,
    and $XDG_* unset. Yields runner instance with its temp dir inside
    `tmp_path` fixture path.
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path_home):
        yield runner


@pytest.fixture
def testdata_path() -> typing.Generator[pathlib.Path, None, None]:
    """Path to local test data directory"""
    yield TESTS_PATH / "testdata"



# SPDX-License-Identifier: Apache-2.0
# Standard

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab.configuration import storage_dirs_exist


@click.group(
    cls=clickext.LazyEntryPointGroup,
    ep_group="instructlab.command.config",
)
@click.pass_context
def config(ctx):
    """Command Group for Interacting with the Config of InstructLab.

    If this is your first time running ilab, it's best to start with `ilab config init` to create the environment.
    """
    ctx.obj = ctx.parent.obj
    if ctx.invoked_subcommand not in {"init"}:
        ctx.obj.ensure_config(ctx)
        if not storage_dirs_exist():
            click.secho(
                "Some ilab storage directories do not exist yet. Please run `ilab config init` before continuing.",
                fg="red",
            )
            raise click.exceptions.Exit(1)
    ctx.default_map = ctx.parent.default_map



# SPDX-License-Identifier: Apache-2.0

# Standard
from typing import Optional, Tuple
import json
import logging
import os
import pathlib
import signal
import subprocess
import sys
import tempfile
import time
import typing

# Third Party
import httpx

# First Party
from instructlab.utils import contains_argument

# Local
from ...client import check_api_base
from ...configuration import get_api_base
from .common import (
    CHAT_TEMPLATE_AUTO,
    CHAT_TEMPLATE_TOKENIZER,
    VLLM,
    Closeable,
    ServerException,
    free_tcp_ipv4_port,
    get_model_template,
    safe_close_all,
    verify_template_exists,
)
from .server import BackendServer, ServerConfig

logger = logging.getLogger(__name__)


class Server(BackendServer):
    def __init__(
        self,
        api_base: str,
        model_family: str,
        model_path: pathlib.Path,
        chat_template: str,
        host: str,
        port: int,
        background: bool = False,
        vllm_args: typing.Iterable[str] | None = (),
        max_startup_attempts: int | None = None,
        log_file: pathlib.Path | None = None,
    ):
        sc = ServerConfig(api_base, log_file)
        super().__init__(model_family, model_path, chat_template, host, port, sc)
        self.api_base = api_base
        self.model_path = model_path
        self.background = background
        self.vllm_args: list[str]
        self.vllm_args = list(vllm_args) if vllm_args is not None else []
        self.process: subprocess.Popen | None = None
        self.max_startup_attempts = max_startup_attempts

    def run(self):
        self.process, files = run_vllm(
            self.host,
            self.port,
            self.model_family,
            self.model_path,
            self.chat_template,
            self.vllm_args,
            self.background,
            log_file=self.config.log_file,
        )
        self.register_resources(files)

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            logger.info("vLLM server terminated by keyboard")
        # pylint: disable=broad-exception-caught
        except BaseException:
            logger.exception("vLLM server terminated")
        finally:
            self.shutdown()

    def create_server_process(self, port: int, background: bool) -> subprocess.Popen:
        server_process, files = run_vllm(
            self.host,
            port,
            self.model_family,
            self.model_path,
            self.chat_template,
            self.vllm_args,
            background=background,
            log_file=self.config.log_file,
        )
        self.register_resources(files)
        return server_process

    def _ensure_server(
        self,
        http_client=None,
        background=True,
        foreground_allowed=False,
    ) -> Tuple[Optional[subprocess.Popen], Optional[str]]:
        """Checks if server is running, if not starts one as a subprocess. Returns the server process
        and the URL where it's available."""

        logger.info(f"Trying to connect to model server at {self.api_base}")
        if check_api_base(self.api_base, http_client):
            return (None, self.api_base)
        port = free_tcp_ipv4_port(self.host)
        logger.debug(f"Using available port {port} for temporary model serving.")

        host_port = f"{self.host}:{port}"
        temp_api_base = get_api_base(host_port)
        vllm_server_process = self.create_server_process(port, background)
        logger.info("Starting a temporary vLLM server at %s", temp_api_base)
        count = 0
        # Each call to check_api_base takes >2s + 2s sleep
        # Default to 120 if not specified (~8 mins of wait time)
        vllm_startup_max_attempts = self.max_startup_attempts or 120
        start_time_secs = time.time()
        while count < vllm_startup_max_attempts:
            count += 1
            # Check if the process is still alive
            if vllm_server_process.poll():
                if foreground_allowed and background:
                    raise ServerException(
                        "vLLM failed to start.  Retry with --enable-serving-output to learn more about the failure."
                    )
                raise ServerException("vLLM failed to start.")
            logger.info(
                "Waiting for the vLLM server to start at %s, this might take a moment... Attempt: %s/%s",
                temp_api_base,
                count,
                vllm_startup_max_attempts,
            )
            if check_api_base(temp_api_base, http_client):
                logger.info("vLLM engine successfully started at %s", temp_api_base)
                break
            if count == vllm_startup_max_attempts:
                logger.info(
                    "Gave up waiting for vLLM server to start at %s after %s attempts",
                    temp_api_base,
                    vllm_startup_max_attempts,
                )
                duration = round(time.time() - start_time_secs, 1)
                shutdown_process(vllm_server_process, 20)
                # pylint: disable=raise-missing-from
                raise ServerException(f"vLLM failed to start up in {duration} seconds")
            time.sleep(2)
        return (vllm_server_process, temp_api_base)

    def run_detached(
        self,
        http_client: httpx.Client | None = None,
        background: bool = True,
        foreground_allowed: bool = False,
        max_startup_retries: int = 0,
    ) -> str:
        for i in range(max_startup_retries + 1):
            try:
                vllm_server_process, api_base = self._ensure_server(
                    http_client=http_client,
                    background=background,
                    foreground_allowed=foreground_allowed,
                )
                self.process = vllm_server_process or self.process
                self.api_base = api_base or self.api_base
                break
            except ServerException as e:
                if i == max_startup_retries:
                    raise e
                logger.info(
                    "vLLM startup failed.  Retrying (%s/%s)",
                    i + 1,
                    max_startup_retries,
                )
                logger.error(e)

        return self.api_base

    def shutdown(self):
        """Shutdown vLLM server"""

        super().shutdown()

        # Needed when a temporary server is started
        if self.process is not None:
            shutdown_process(self.process, 20)
            self.process = None

    def get_backend_type(self):
        return VLLM


def format_template(model_family: str, model_path: pathlib.Path) -> str:
    template, eos_token, bos_token = get_model_template(model_family, model_path)

    prefix = ""
    if eos_token:
        prefix = '{{% set eos_token = "{}" %}}\n'.format(eos_token)
    if bos_token:
        prefix = '{}{{% set bos_token = "{}" %}}\n'.format(prefix, bos_token)

    return prefix + template


def get_argument(prefix: str, args: typing.List[str]):
    # Return last value in args for either --foo value or --foo=value
    # Returns True if flag --foo is provided with no value

    args_len = len(args)
    # Traverse the args in reverse (args_len-1 to 0)
    for i in range(args_len - 1, -1, -1):
        s = args[i]
        if s == prefix:
            # Case: --foo value or --foo (with no value)
            if i == args_len - 1:
                # --foo is the last entry, must be a flag
                return True
            next_arg = args[i + 1]
            if next_arg.startswith("-"):
                # No value provided, must be a flag
                return True
            # The entry after is the value
            return next_arg
        v = prefix + "="
        if s.startswith(v):
            # Case: --foo=value
            # Return everything after prefix=
            return s[len(v) :]
    return None


def create_tmpfile(data: str):
    # pylint: disable=consider-using-with
    file = tempfile.NamedTemporaryFile()
    file.write(data.encode("utf-8"))
    file.flush()
    return file


def run_vllm(
    host: str,
    port: int,
    model_family: str,
    model_path: pathlib.Path,
    chat_template: str,
    vllm_args: list[str],
    background: bool,
    log_file: pathlib.Path | None = None,
) -> typing.Tuple[subprocess.Popen, list[Closeable]]:
    """
    Start an OpenAI-compatible server with vLLM.

    Args:
        host          (str):          host to run server on
        port          (int):          port to run server on
        model_family  (str):          Family the model belongs to, used with 'auto' chat templates
        model_path    (Path):         The path to the model file
        chat_template (str):          Chat template to use when serving the model
                                         'auto' (default) automatically determines a template\n
                                         'tokenizer' uses the model provided template\n
                                         (path) specifies a template file location to load from.

        vllm_args     (list of str):  Specific arguments to pass into vllm.
                                        Example: ["--dtype", "auto", "--enable-lora"]
        background (bool):            Whether the stdout and stderr vLLM should be sent to /dev/null (True)
                                      or stay in the foreground(False).
        log_file (Path):              File to write stdout and stderr
    Returns:
        tuple: A tuple containing two values:
            vllm_process (subprocess.Popen): process of the vllm server
            tmp_files: a list of temporary files necessary to launch the process

    """
    vllm_process = None
    vllm_cmd, tmp_files = build_vllm_cmd(
        host, port, model_family, model_path, chat_template, vllm_args
    )

    logger.debug(f"vLLM serving command is: {vllm_cmd}")

    vllm_env = os.environ.copy()
    # Reset vllm logging to the default (enabled)
    vllm_env.pop("VLLM_CONFIGURE_LOGGING", None)

    try:
        # Note: start_new_session=True is needed to create a process group which will later be used
        # on shutdown. The new process will not be a child of the current process group. Instead, it
        # will be the leader of a new session and process group.
        if background:
            if log_file:
                # write both stdout and stderr to the log file in append mode
                with log_file.open("a", encoding="utf-8") as f:
                    vllm_process = subprocess.Popen(
                        args=vllm_cmd,
                        stdout=f,
                        stderr=f,
                        start_new_session=True,
                        env=vllm_env,
                    )
            else:
                vllm_process = subprocess.Popen(
                    args=vllm_cmd,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    env=vllm_env,
                    start_new_session=True,
                )
        else:
            # pylint: disable=consider-using-with
            vllm_process = subprocess.Popen(
                args=vllm_cmd,
                env=vllm_env,
                start_new_session=True,
            )

        api_base = get_api_base(f"{host}:{port}")
        logger.info("vLLM starting up on pid %s at %s", vllm_process.pid, api_base)

    except:
        safe_close_all(tmp_files)
        raise

    return vllm_process, tmp_files


def is_bnb_quantized(model_path: pathlib.Path) -> bool:
    """
    Check if provided model has quantization config with bitsandbytes specified.

    Args:
        model_path    (Path):         The path to the model files

    Returns:
        bool: Whether or not the model has been quantized via bitsandbytes
    """
    config_json = model_path / "config.json"
    if not model_path.is_dir() or not config_json.is_file():
        return False

    with config_json.open(encoding="utf-8") as f:
        config = json.load(f)

    return bool(
        config.get("quantization_config", {}).get("quant_method", "") == "bitsandbytes"
    )


def build_vllm_cmd(
    host: str,
    port: int,
    model_family: str,
    model_path: pathlib.Path,
    chat_template: str,
    vllm_args: list[str],
) -> typing.Tuple[list[str], list[Closeable]]:
    """
    Build the vLLM command to run the server.

    Args:
        host          (str):          host to run server on
        port          (int):          port to run server on
        model_family  (str):          Family the model belongs to, used with 'auto' chat templates
        model_path    (Path):         The path to the model file
        chat_template (str):          Chat template to use when serving the model
                                         'auto' (default) automatically determines a template\n
                                         'tokenizer' uses the model provided template\n
                                         (path) specifies a template file location to load from.

        vllm_args     (list of str):  Specific arguments to pass into vllm.
                                        Example: ["--dtype", "auto", "--enable-lora"]

    Returns:
        tuple: A tuple containing two values:
            vllm_cmd (list of str): The command to run the vllm server
            tmp_files: a list of temporary files necessary to launch the process
    """
    vllm_cmd = [sys.executable, "-m", "vllm.entrypoints.openai.api_server"]
    tmp_files = []

    if not contains_argument("--host", vllm_args):
        vllm_cmd.extend(["--host", host])

    if not contains_argument("--port", vllm_args):
        vllm_cmd.extend(["--port", str(port)])

    if not contains_argument("--model", vllm_args):
        vllm_cmd.extend(["--model", os.fspath(model_path)])

    if not contains_argument("--chat-template", vllm_args):
        if chat_template == CHAT_TEMPLATE_AUTO:
            # For auto templates, the build-in in-memory template
            # needs to be written to a temporary file so that vLLM
            # can read it
            template = format_template(model_family, model_path)
            file = create_tmpfile(template)
            tmp_files.append(file)
            vllm_cmd.extend(["--chat-template", file.name])
        elif chat_template != CHAT_TEMPLATE_TOKENIZER:
            # In this case the value represents a template file path
            # Pass it directly to vllm
            path = pathlib.Path(chat_template)
            verify_template_exists(path)
            vllm_cmd.extend(["--chat-template", chat_template])

    # Auto-detect whether model is quantized w/ bitsandbytes and add potentially missing args
    quant_arg_present = contains_argument("--quantization", vllm_args)
    load_arg_present = contains_argument("--load-format", vllm_args)
    eager_arg_present = contains_argument("--enforce-eager", vllm_args)
    if not (quant_arg_present and load_arg_present and eager_arg_present):
        if is_bnb_quantized(model_path):
            if not quant_arg_present:
                vllm_cmd.extend(["--quantization", "bitsandbytes"])
            if not load_arg_present:
                vllm_cmd.extend(["--load-format", "bitsandbytes"])
            # Currently needed to retain generation quality w/ 4-bit bnb + vLLM (bypass graph bug)
            if not eager_arg_present:
                vllm_cmd.append("--enforce-eager")

    # Force multiprocessing for distributed serving, vLLM will try "ray" if it's installed but we do
    # not support it (yet?). We don't install Ray but we might end up running on systems that have it,
    # so let's make sure we use multiprocessing.
    if not contains_argument("--distributed-executor-backend", vllm_args):
        vllm_cmd.extend(["--distributed-executor-backend", "mp"])

    vllm_cmd.extend(vllm_args)

    return vllm_cmd, tmp_files


def shutdown_process(process: subprocess.Popen, timeout: int) -> None:
    """
    Shuts down a process

    Sends SIGINT and then after a timeout if the process still is not terminated sends
    a SIGKILL. Finally, a SIGKILL is sent to the process group in case any child processes
    weren't cleaned up.

    Args:
        process (subprocess.Popen): process of the vllm server

    Returns:
        Nothing
    """
    # vLLM responds to SIGINT by shutting down gracefully and reaping the children
    logger.debug(f"Sending SIGINT to vLLM server PID {process.pid}")
    process_group_id = os.getpgid(process.pid)
    process.send_signal(signal.SIGINT)
    try:
        logger.debug("Waiting for vLLM server to shut down gracefully")
        process.wait(timeout)
    except subprocess.TimeoutExpired:
        logger.debug(
            f"Sending SIGKILL to vLLM server since timeout ({timeout}s) expired"
        )
        process.kill()

    # Attempt to cleanup any remaining child processes
    # Make sure process_group is legit (> 1) before trying
    if process_group_id and process_group_id > 1:
        try:
            os.killpg(process_group_id, signal.SIGKILL)
            logger.debug("Sent SIGKILL to vLLM process group")
        except ProcessLookupError:
            logger.debug("Nothing left to clean up with the vLLM process group")
    else:
        logger.debug("vLLM process group id not found")

    # Various facilities of InstructLab rely on multiple successive start/stop
    # cycles. Since vLLM relies on stable VRAM for estimating capacity, residual
    # reclamation activity can lead to crashes on start. To prevent this add a
    # short delay (typically ~ 10 seconds, max 30) to verify stability.
    #
    # Ideally a future enhancement would be contributed to vLLM to more gracefully
    # handle this condition.
    wait_for_stable_vram(get_max_stable_vram_wait(30))


def wait_for_stable_vram(timeout: int):
    logger.info("Waiting for GPU VRAM reclamation...")
    supported, stable = wait_for_stable_vram_cuda(timeout)
    if not supported:
        # TODO add support for intel
        time.sleep(timeout)
        return
    if not stable:
        # Only for debugging since recovery is likely after additional start delay
        logger.debug(
            "GPU VRAM did not stabilize during max timeout (%d seconds)", timeout
        )


def get_max_stable_vram_wait(timeout: int) -> int:
    # Internal env variable for CI adjustment / disablement
    env_name = "ILAB_MAX_STABLE_VRAM_WAIT"
    wait_str = os.getenv(env_name)
    wait = timeout
    try:
        if wait_str:
            wait = int(wait_str)
    except ValueError:
        logger.warning(
            'Ignoring invalid timeout value for %s ("%s")', env_name, wait_str
        )

    return wait


def wait_for_stable_vram_cuda(timeout: int) -> Tuple[bool, bool]:
    if timeout == 0:
        logger.debug("GPU vram stability check disabled with 0 max wait value")
        return True, True

    # Third Party
    import torch

    # Fallback to a constant sleep if we don't have support for the device
    if not torch.cuda.is_available():
        return False, False
    start_time = time.monotonic()
    stable_samples = 0
    last_free = 0
    try:
        while True:
            free_memory = 0
            try:
                # TODO In the future this should be enhanced to better handle
                # GPU partitioning. However, to do so will require that serve
                # assign specific GPUs to vLLM, so that the same device pool is
                # analyzed.
                for i in range(torch.cuda.device_count()):
                    device = torch.device(f"cuda:{i}")
                    free_memory += torch.cuda.mem_get_info(device)[0]
            except Exception:  # pylint: disable=broad-exception-caught
                logger.warning(
                    "Could not determine CUDA memory, falling back to general sleep"
                )
                return False, False

            # Wait for free memory to stop growing indicating the release of
            # vram after vLLM shutdown is complete. Wait for 5 successful
            # samples where this is true, but ignore any spurious readings that
            # occur between those samples. In the future we may be able to
            # optimize this by checking a few strictly successive samples.
            if free_memory <= last_free:
                stable_samples += 1
                logger.debug(
                    "GPU free vram stable (stable count %d, free %d, last free %d)",
                    stable_samples,
                    free_memory,
                    last_free,
                )
                if stable_samples > 5:
                    logger.debug(
                        "Successful sample recorded, (stable count %d, free %d, last free %d)",
                        stable_samples,
                        free_memory,
                        last_free,
                    )
                    return True, True
            else:
                if last_free != 0:
                    logger.debug(
                        "GPU free vram still growing (free %d, last free %d)",
                        free_memory,
                        last_free,
                    )

            if time.monotonic() - start_time > timeout:
                return True, False

            last_free = free_memory
            time.sleep(1)
    finally:
        try:
            torch.cuda.empty_cache()
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.debug("Could not free cuda cache: %s", e)



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
from typing import List, Optional
import shutil

# Third Party
import git

TEST_VALID_COMPOSITIONAL_SKILL_YAML = """created_by: rafael-vasquez
version: 1
seed_examples:
- answer: "Sure thing!"
  context: "This is a valid YAML."
  question: "Can you help me debug this failing unit test?"
- answer: "answer2"
  context: "context2"
  question: "question2"
- answer: "answer3"
  context: "context3"
  question: "question3"
- answer: "answer4"
  context: "context4"
  question: "question4"
- answer: "answer5"
  context: "context5"
  question: "question5"
task_description: 'This is a task'
"""


class MockTaxonomy:
    INIT_COMMIT_FILE = "README.md"

    def __init__(self, path: Path) -> None:
        self.root = path
        self._repo = git.Repo.init(path, initial_branch="main")
        with open(path / self.INIT_COMMIT_FILE, "wb"):
            pass
        self._repo.index.add([self.INIT_COMMIT_FILE])
        self._repo.index.commit("Initial commit")

    @property
    def untracked_files(self) -> List[str]:
        """List untracked files in the repository"""
        return self._repo.untracked_files

    def create_untracked(self, rel_path: str, contents: Optional[bytes] = None) -> Path:
        """Create a new untracked file in the repository.

        Args:
            rel_path (str): Relative path (from repository root) to the file.
            contents (bytes): (optional) Byte string to be written to the file.
        Returns:
            file_path: The path to the created file.
        """
        taxonomy_path = Path(rel_path)
        assert not taxonomy_path.is_absolute()
        file_path = self.root.joinpath(taxonomy_path)
        file_path.parent.mkdir(exist_ok=True, parents=True)
        if not contents:
            assert taxonomy_path.parts[0] == "compositional_skills"
            file_path.write_text(TEST_VALID_COMPOSITIONAL_SKILL_YAML, encoding="utf-8")
        else:
            file_path.write_bytes(contents)
        return file_path

    def add_tracked(self, rel_path: str) -> Path:
        """Add a new tracked file to the repository (and commits it).

        Args:
            rel_path (str): Relative path (from repository root) to the file.
        Returns:
            file_path: The path to the added file.
        """
        file_path = self.create_untracked(rel_path)
        self._repo.index.add([rel_path])
        self._repo.index.commit("new commit")
        return file_path

    def remove_file(self, rel_path: str) -> None:
        """Remove a file in the repository (tracked or not)

        Args:
            rel_path (str): Relative path (from repository root) to the file.
        """
        self.root.joinpath(rel_path).unlink(missing_ok=True)

    def teardown(self) -> None:
        """Recursively remove the temporary repository and all of its
        subdirectories and files.
        """
        shutil.rmtree(self.root)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.teardown()



# SPDX-License-Identifier: Apache-2.0

# Standard
from typing import Any
from unittest.mock import patch
import logging
import os
import pathlib
import shutil
import typing

# Third Party
from pydantic import BaseModel, Field
from xdg_base_dirs import xdg_cache_home, xdg_data_home
import pydantic
import pytest
import yaml

# First Party
from instructlab import configuration as config
from instructlab.clickext import ConfigOption, get_default_and_description
from instructlab.configuration import DEFAULTS
from instructlab.log import configure_logging


class TestConfig:
    def _assert_defaults(self, cfg: config.Config):
        # redefine defaults here instead of relyin on those in configuration.DEFAULTS
        # to catch any errors if we are doing things incorrectly over there
        package_name = "instructlab"
        internal_dirname = "internal"
        cache_dir = os.path.join(xdg_cache_home(), package_name)
        data_dir = os.path.join(xdg_data_home(), package_name)
        default_chat_model = f"{cache_dir}/models/merlinite-7b-lab-Q4_K_M.gguf"
        default_sdg_model = f"{cache_dir}/models/mistral-7b-instruct-v0.2.Q4_K_M.gguf"

        assert cfg.general is not None
        assert cfg.version is not None
        assert cfg.general.log_level == "INFO"
        assert cfg.general.debug_level == 0

        assert cfg.chat is not None
        assert cfg.chat.model == default_chat_model
        assert not cfg.chat.vi_mode
        assert cfg.chat.visible_overflow
        assert cfg.chat.context == "default"
        assert cfg.chat.session is None
        assert cfg.chat.logs_dir == f"{data_dir}/chatlogs"
        assert not cfg.chat.greedy_mode

        assert cfg.evaluate is not None

        assert cfg.general is not None
        assert cfg.general.log_level == "INFO"

        assert cfg.generate is not None
        assert cfg.generate.teacher.model_path == default_sdg_model
        assert cfg.generate.teacher.llama_cpp is not None
        assert cfg.generate.teacher.llama_cpp.gpu_layers == -1
        assert cfg.generate.teacher.llama_cpp.max_ctx_size == 4096
        assert cfg.generate.teacher.llama_cpp.llm_family == ""
        assert cfg.generate.teacher.vllm is not None
        assert cfg.generate.teacher.vllm.vllm_args == []
        assert cfg.generate.teacher.host_port == "127.0.0.1:8000"
        assert cfg.generate.teacher.backend is None
        assert cfg.generate.teacher.chat_template is None
        assert cfg.generate.pipeline == "full"
        assert cfg.generate.model == default_sdg_model
        assert cfg.generate.taxonomy_path == f"{data_dir}/taxonomy"
        assert cfg.generate.taxonomy_base == "origin/main"
        assert cfg.generate.num_cpus == 10
        assert cfg.generate.sdg_scale_factor == 30
        assert cfg.generate.chunk_word_count == 1000
        assert cfg.generate.output_dir == f"{data_dir}/datasets"
        assert (
            cfg.generate.seed_file == f"{data_dir}/{internal_dirname}/seed_tasks.json"
        )

        assert cfg.serve is not None
        assert cfg.serve.model_path == default_chat_model
        assert cfg.serve.llama_cpp is not None
        assert cfg.serve.llama_cpp.gpu_layers == -1
        assert cfg.serve.llama_cpp.max_ctx_size == 4096
        assert cfg.serve.llama_cpp.llm_family == ""
        assert cfg.serve.vllm is not None
        assert cfg.serve.vllm.vllm_args == []
        assert cfg.serve.host_port == "127.0.0.1:8000"
        assert cfg.serve.backend is None
        assert cfg.serve.chat_template is None

    def _assert_model_defaults(self, cfg):
        package_name = "instructlab"
        cache_dir = os.path.join(xdg_cache_home(), package_name)
        default_chat_model = f"{cache_dir}/models/merlinite-7b-lab-Q4_K_M.gguf"

        assert cfg.chat is not None
        assert cfg.chat.model == default_chat_model

        assert cfg.evaluate is not None
        assert cfg.evaluate.base_model == "instructlab/granite-7b-lab"

        assert cfg.generate is not None
        assert cfg.generate.model == DEFAULTS.DEFAULT_TEACHER_MODEL

        assert cfg.serve is not None
        assert cfg.serve.model_path == DEFAULTS.DEFAULT_CHAT_MODEL

        assert cfg.train is not None
        assert cfg.train.model_path == "instructlab/granite-7b-lab"

    def test_default_config(self, cli_runner):  # pylint: disable=unused-argument
        cfg = config.get_default_config()
        assert cfg is not None
        self._assert_defaults(cfg)
        self._assert_model_defaults(cfg)

    def test_cfg_auto_fill(self, tmp_path_home):
        config_path = tmp_path_home / "config.yaml"
        with open(config_path, "w", encoding="utf-8") as config_file:
            config_file.write(
                """general:
  log_level: INFO
"""
            )
        cfg = config.read_config(config_path)
        self._assert_defaults(cfg)
        self._assert_model_defaults(cfg)

    def test_cfg_auto_fill_with_large_config(self, tmp_path_home):  # pylint: disable=unused-argument
        config_path = tmp_path_home / "config.yaml"
        with open(config_path, "w", encoding="utf-8") as config_file:
            config_file.write(
                """general:
  log_level: INFO
generate:
  pipeline: simple
  teacher:
    model_path: models/granite-7b-lab-Q4_K_M.gguf
    chat_template: tokenizer
    llama_cpp:
      gpu_layers: 1
      max_ctx_size: 2048
      llm_family: ''
    vllm:
      gpus: 8
"""
            )
        cfg = config.read_config(config_path)
        # make sure the generate cfg passed is preserved
        assert cfg.generate.teacher.llama_cpp.max_ctx_size == 2048
        self._assert_model_defaults(cfg)

    def test_validate_log_level_invalid(self):
        cfg = config.get_default_config()
        with pytest.raises(ValueError):
            cfg.general.validate_log_level("INVALID")

    def test_validate_log_level_valid(self):
        cfg = config.get_default_config()
        assert cfg.general.validate_log_level("DEBUG") == "DEBUG"
        assert cfg.general.validate_log_level("INFO") == "INFO"
        assert cfg.general.validate_log_level("WARNING") == "WARNING"
        assert cfg.general.validate_log_level("WARN") == "WARN"
        assert cfg.general.validate_log_level("FATAL") == "FATAL"
        assert cfg.general.validate_log_level("CRITICAL") == "CRITICAL"
        assert cfg.general.validate_log_level("ERROR") == "ERROR"
        assert cfg.general.validate_log_level("NOTSET") == "NOTSET"

    def test_expand_paths(self, tmp_path_home):
        config_path = tmp_path_home / "config.yaml"
        with open(config_path, "w", encoding="utf-8") as config_file:
            config_file.write(
                """\
version: 1.0.0
chat:
  model: $HOME/.cache/instructlab/models/granite-7b-lab-Q4_K_M.gguf
generate:
  model: ~/.cache/instructlab/models/granite-7b-lab-Q4_K_M.gguf
  taxonomy_base: upstream/main
  taxonomy_path: mytaxonomy
  teacher:
    model_path: ~/.cache/instructlab/models/granite-7b-lab-Q4_K_M.gguf
    chat_template: tokenizer
    llama_cpp:
      gpu_layers: 1
      max_ctx_size: 2048
      llm_family: ''
    vllm:
      gpus: 8
serve:
  model_path: models/granite-7b-lab-Q4_K_M.gguf
  chat_template: tokenizer
  llama_cpp:
    gpu_layers: 1
    max_ctx_size: 2048
    llm_family: ''
  vllm:
    gpus: 8
    vllm_args:
       - --qlora-adapter-name-or-path=$HOME/qlora-adapter-name-or-path
"""
            )
        cfg = config.read_config(config_path)
        assert cfg.chat.model.startswith("/")
        assert cfg.chat.model == os.path.expandvars(
            "$HOME/.cache/instructlab/models/granite-7b-lab-Q4_K_M.gguf"
        )
        assert cfg.generate.model.startswith("/")
        assert cfg.generate.model == os.path.expanduser(
            "~/.cache/instructlab/models/granite-7b-lab-Q4_K_M.gguf"
        )
        # Validate multi level dict
        assert cfg.generate.teacher.model_path.startswith("/")
        assert cfg.generate.teacher.model_path == os.path.expanduser(
            "~/.cache/instructlab/models/granite-7b-lab-Q4_K_M.gguf"
        )
        assert cfg.serve.vllm.vllm_args[0] == os.path.expandvars(
            "--qlora-adapter-name-or-path=$HOME/qlora-adapter-name-or-path"
        )

    def test_validate_log_format_invalid(self):
        cfg = config.get_default_config()
        with pytest.raises(ValueError):
            cfg.general.validate_log_format("INVALID")

    def test_get_model_family(self):
        good_cases = {
            # two known families
            "merlinite": "merlinite",
            "mixtral": "mixtral",
            # case insensitive
            "MERLINiTe": "merlinite",
            # mapping granite to merlinite
            "granite": "merlinite",
            # default empty value of model_family will use name of model in model path to guess model_family
            "": "merlinite",
        }
        bad_cases = [
            # unknown family
            "unknown",
        ]
        for model_name, expected_family in good_cases.items():
            model_path = os.path.join("models", f"{model_name}-7b-lab-Q4_K_M.gguf")
            assert config.get_model_family(model_name, model_path) == expected_family
            assert config.get_model_family(None, model_path) == expected_family
        for model_name in bad_cases:
            model_path = os.path.join("models", f"{model_name}-7b-lab-Q4_K_M.gguf")
            with pytest.raises(config.ConfigException):
                config.get_model_family(model_name, model_path)

    def test_config_modified_settings(self, tmp_path_home):
        config_path = tmp_path_home / "config.yaml"
        with open(config_path, "w", encoding="utf-8") as config_file:
            config_file.write(
                """\
version: 1.0.0
chat:
  model: models/granite-7b-lab-Q4_K_M.gguf
generate:
  pipeline: simple
  model: models/granite-7b-lab-Q4_K_M.gguf
  taxonomy_base: upstream/main
  taxonomy_path: mytaxonomy
  teacher:
    model_path: models/granite-7b-lab-Q4_K_M.gguf
    chat_template: tokenizer
    llama_cpp:
      gpu_layers: 1
      max_ctx_size: 2048
      llm_family: ''
    vllm:
      gpus: 8
serve:
  model_path: models/granite-7b-lab-Q4_K_M.gguf
  chat_template: tokenizer
  llama_cpp:
    gpu_layers: 1
    max_ctx_size: 2048
    llm_family: ''
  vllm:
    gpus: 8
    vllm_args:
       - --dtype=auto
       - --enable-lora
evaluate:
  base_model: instructlab/granite-7b-lab
  gpus: 1
  mmlu:
    few_shots: 2
    batch_size: auto
  mmlu_branch:
    tasks_dir: /path/to/sdg
  mt_bench:
    judge_model: prometheus
    output_dir: /dir/to/output
    max_workers: auto
  mt_bench_branch:
    taxonomy_path: taxonomy
general:
  log_level: DEBUG
"""
            )
        cfg = config.read_config(config_path)
        assert cfg is not None
        assert cfg.chat.model == "models/granite-7b-lab-Q4_K_M.gguf"
        assert cfg.generate.pipeline == "simple"
        assert cfg.generate.model == "models/granite-7b-lab-Q4_K_M.gguf"
        assert cfg.serve.llama_cpp.gpu_layers == 1
        assert cfg.serve.llama_cpp.max_ctx_size == 2048
        assert cfg.serve.chat_template == "tokenizer"
        assert cfg.serve.vllm.vllm_args == [
            "--dtype=auto",
            "--enable-lora",
        ]
        assert cfg.general.log_level == "DEBUG"
        assert cfg.general.debug_level == 1


@pytest.mark.parametrize(
    "log_level,debug_level,root,instructlab,openai_httpx",
    [
        ("INFO", 0, logging.INFO, logging.INFO, logging.ERROR),
        ("DEBUG", 1, logging.INFO, logging.DEBUG, logging.ERROR),
        ("DEBUG", 2, logging.DEBUG, logging.DEBUG, logging.DEBUG),
        ("ERROR", 0, logging.ERROR, logging.ERROR, logging.ERROR),
    ],
)
def test_logging(log_level, debug_level, root, instructlab, openai_httpx):
    configure_logging(log_level=log_level, debug_level=debug_level, fmt="%(message)s")
    assert logging.getLogger("root").getEffectiveLevel() == root
    assert logging.getLogger("instructlab").getEffectiveLevel() == instructlab
    assert logging.getLogger("openai").getEffectiveLevel() == openai_httpx
    assert logging.getLogger("httpx").getEffectiveLevel() == openai_httpx
    assert logging.getLogger("root").handlers[0].formatter._fmt == "%(message)s"


@patch.multiple(
    config.DEFAULTS,
    _cache_home="/cache/instructlab",
    _config_dir="/config/instructlab",
    _data_dir="/data/instructlab",
)
def test_compare_default_config_testdata(
    testdata_path: pathlib.Path, tmp_path: pathlib.Path, regenerate_testdata: bool
):
    assert config.DEFAULTS.CHECKPOINTS_DIR == "/data/instructlab/checkpoints"
    saved_file = testdata_path / "default_config.yaml"
    current_file = tmp_path / "current_config.yaml"

    current_cfg = config.get_default_config()
    # roundtrip to verify serialization and de-serialization
    config.write_config(current_cfg, str(current_file))
    with current_file.open(encoding="utf-8") as yamlfile:
        current_content = yaml.safe_load(yamlfile)

    if regenerate_testdata:
        shutil.copy(current_file, saved_file)

    with saved_file.open(encoding="utf-8") as yamlfile:
        saved_content = yaml.safe_load(yamlfile)

    assert current_content == saved_content, (
        "current and expected configs are different. If the change was "
        "intentional, run 'make regenerate-testdata' and commit the "
        "updated test data."
    )


def test_all_config_options_have_description():
    cfg = config.get_default_config()

    def loop_through_fields(
        cfg: pydantic.BaseModel,
    ) -> typing.List[str]:
        fields_without_description = []
        for field_name, field in cfg.model_fields.items():
            value = getattr(cfg, field_name)
            description = field.description

            if isinstance(value, pydantic.BaseModel):
                fields_without_description.extend(loop_through_fields(value))

            if not description:
                fields_without_description.append(field_name)

        return fields_without_description

    fields_without_description = loop_through_fields(cfg)
    assert not fields_without_description, f"Fields without description: {fields_without_description}, update 'configuration.py' Field with missing descriptions."


@pytest.mark.parametrize(
    "lora_quantize_dtype,additional_args,raises_exception",
    [
        ("nf4", {}, False),
        (None, {}, False),
        ("valid-for-cli-but-not-for-training-library", {}, False),
        ("nf4", {"lora_alpha": 32}, False),
    ],
)
def test_read_train_profile(
    lora_quantize_dtype: str | None,
    additional_args: dict[str, Any] | None,
    raises_exception: bool,
    tmp_path_home,
):
    # define a profile with yaml
    profile_data = {
        "model_path": "/path/to/model",
        "data_path": "/path/to/data",
        "ckpt_output_dir": "/path/to/checkpoints",
        "data_output_dir": "/dev/shm",
        "max_seq_len": 4096,
        "max_batch_len": 60_000,
        "num_epochs": 10,
        "effective_batch_size": 3840,
        "save_samples": 250_000,
        "deepspeed_cpu_offload_optimizer": True,
        "lora_rank": 4,
        "lora_quantize_dtype": lora_quantize_dtype,
        "is_padding_free": False,
        "nproc_per_node": 8,
        "additional_args": additional_args,
    }
    train_profile = pathlib.Path(tmp_path_home, "profile.yaml")
    with open(train_profile, "w", encoding="utf-8") as outfile:
        yaml.dump(profile_data, outfile)

    if raises_exception:
        with pytest.raises(config.ConfigException):
            config.read_train_profile(train_profile)
    else:
        result = config.read_train_profile(train_profile)
        assert result is not None
        for k, v in result.dict().items():
            if k not in profile_data:
                continue
            if profile_data[k] is not None:
                assert v == profile_data[k]


class NestedConfig(BaseModel):
    nested_field: str = Field(
        default="nested_default", description="Nested field description"
    )


class Config(BaseModel):
    field: str = Field(default="default_value", description="Field description")
    nested: NestedConfig = NestedConfig()


@pytest.mark.parametrize(
    "cfg, config_identifier, expected_description, expected_default_value, expect_error",
    [
        (Config(), ["field"], "Field description", "default_value", False),
        (
            Config(),
            ["nested", "nested_field"],
            "Nested field description",
            "nested_default",
            False,
        ),
        (Config(), ["non_existent_field"], None, None, True),
        (Config(), ["nested", "non_existent_field"], None, None, True),
    ],
)
def test_get_default_and_description(
    cfg, config_identifier, expected_description, expected_default_value, expect_error
):
    if expect_error:
        with pytest.raises(ValueError):
            get_default_and_description(cfg, config_identifier)
    else:
        description, default_value = get_default_and_description(cfg, config_identifier)
        assert description == expected_description
        assert default_value == expected_default_value


@pytest.mark.parametrize(
    "param_decls, help_text, show_default, should_raise",
    [
        (
            ["--test"],
            "This is a test help message",
            False,
            True,
        ),  # Should raise ValueError since there is help text but show_default=False
        (["--test"], None, False, False),  # Should not raise ValueError
        (
            ["--test"],
            None,
            True,
            True,
        ),  # Should raise ValueError with show_default=True
    ],
)
def test_click_option(param_decls, help_text, show_default, should_raise):
    if should_raise:
        with pytest.raises(ValueError):
            ConfigOption(
                param_decls=param_decls,
                help=help_text,
                show_default=show_default,
            )
    else:
        try:
            ConfigOption(
                param_decls=param_decls,
                help=help_text,
                show_default=show_default,
            )
        except ValueError as e:
            pytest.fail(f"ConfigOption raised an unexpected ValueError: {e}")



# SPDX-License-Identifier: Apache-2.0

# Standard
import logging
import os

# Third Party
from git import GitError
import click
import yaml

# First Party
from instructlab import clickext

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--taxonomy-path",
    type=click.Path(),
    help="Path to where the taxonomy is stored locally.",
    show_default="Default taxonomy location in the instructlab data directory.",
)
@click.option(
    "--taxonomy-base",
    help="Base git-ref to use for taxonomy, use 'empty' for the entire repo contents.",
)
@click.option(
    "--yaml-rules",
    type=click.Path(),
    default=None,
    help="Custom rules file for YAML linting.",
)
@click.option(
    "--quiet",
    is_flag=True,
    help="Suppress all output. Call returns 0 if check passes, 1 otherwise.",
)
@click.pass_context
@clickext.display_params
def diff(ctx, taxonomy_path, taxonomy_base, yaml_rules, quiet):
    """
    Lists taxonomy files that have changed since <taxonomy-base>
    and checks that taxonomy is valid. Similar to 'git diff <ref>'.
    """
    # pylint: disable=import-outside-toplevel
    # Third Party
    from instructlab.schema.taxonomy import TaxonomyReadingException

    # Local
    from ..utils import get_taxonomy, get_taxonomy_diff, validate_taxonomy

    # load defaults from config ctx obj if not specified via CLI arg
    if not taxonomy_base:
        taxonomy_base = ctx.obj.config.generate.taxonomy_base
    if not taxonomy_path:
        taxonomy_path = ctx.obj.config.generate.taxonomy_path

    # get all new or changed taxonomy files to be validated and output to user
    if not quiet:
        is_file = os.path.isfile(taxonomy_path)
        if is_file:  # taxonomy_path is file
            click.echo(taxonomy_path)
        else:  # taxonomy_path is dir
            if taxonomy_base == "empty":
                # Gather all the yamls - equivalent to a diff against "the null tree"
                taxonomy_files = get_taxonomy(taxonomy_path)
            else:
                try:
                    # Gather the new or changed YAMLs using git diff, including untracked files
                    taxonomy_files = get_taxonomy_diff(taxonomy_path, taxonomy_base)
                except (TaxonomyReadingException, GitError) as exc:
                    click.secho(
                        f"Reading taxonomy failed with the following error: {exc}",
                        fg="red",
                    )
                    raise SystemExit(1) from exc
            for f in taxonomy_files:
                click.echo(f)

    # validate new or changed taxonomy files
    try:
        validate_taxonomy(taxonomy_path, taxonomy_base, yaml_rules)
    except (TaxonomyReadingException, yaml.YAMLError) as exc:
        if not quiet:
            click.secho(
                f"Reading taxonomy failed with the following error: {exc}",
                fg="red",
            )
        raise SystemExit(1) from exc
    if not quiet:
        click.secho(
            f"Taxonomy in {taxonomy_path} is valid :)",
            fg="green",
        )



# Standard
from pathlib import Path
import logging
import os
import pathlib
import shutil

# First Party
from instructlab import utils
from instructlab.configuration import DEFAULTS

# pylint: disable=ungrouped-imports

logger = logging.getLogger(__name__)


def simple_train(
    model_path,
    skip_preprocessing,
    skip_quantize,
    gguf_model_path,
    tokenizer_dir,
    data_path,
    input_dir,
    ckpt_output_dir,
    iters,
    local,
    num_epochs,
    device,
    four_bit_quant,
):
    effective_data_dir: pathlib.Path = Path(
        data_path if data_path else DEFAULTS.DATASETS_DIR
    )

    train_file = effective_data_dir / "train_gen.jsonl"
    test_file = effective_data_dir / "test_gen.jsonl"

    # NOTE: If given a data_dir, input-dir is ignored in favor of existing!
    if not data_path or data_path.strip() == DEFAULTS.DATASETS_DIR:
        data_path = str(effective_data_dir)
        if not os.path.exists(input_dir):
            raise OSError(
                f"Could not read directory: {input_dir}",
            )

        try:
            os.makedirs(data_path, exist_ok=True)
        except OSError as exc:
            raise OSError(f"Could not create data dir: {exc}") from exc

        # generated input files reverse sorted by modification time
        def get_files(directory: str, pattern: str) -> list[str]:
            return sorted(
                [str(p) for p in Path(directory).glob(pattern)],
                key=os.path.getmtime,
                reverse=True,
            )

        # ignore the test_file and train_file to prevent it from being copied back onto itself
        # see: https://github.com/instructlab/instructlab/pull/1685
        test_files = [
            f
            for f in get_files(input_dir, "test_*")
            if os.path.basename(f) != os.path.basename(test_file)
        ]
        train_files = [
            f
            for f in get_files(input_dir, "train_*")
            if os.path.basename(f) != os.path.basename(train_file)
        ]

        if not train_files or not test_files:
            raise FileNotFoundError(
                f"{input_dir} does not contain training or test files, did you run `ilab data generate`?",
            )
        if len(train_files) > 1 or len(test_files) > 1:
            logger.warning(
                "Found multiple files from `ilab data generate`. Using the most recent generation.",
            )
        # The first file is latest
        logger.debug("train_file=%s", train_files[0])
        logger.debug("test_file=%s", test_files[0])
        shutil.copy(train_files[0], train_file)
        shutil.copy(test_files[0], test_file)

    if utils.is_macos_with_m_chip():
        # Local
        from ..mlx_explore.gguf_convert_to_mlx import load
        from ..mlx_explore.utils import fetch_tokenizer_from_hub
        from ..train.lora_mlx.convert import convert_between_mlx_and_pytorch
        from ..train.lora_mlx.lora import load_and_train
        from ..train.lora_mlx.make_data import make_data

        if not skip_preprocessing:
            try:
                make_data(data_dir=data_path)
            except FileNotFoundError as exc:
                raise FileNotFoundError(
                    f"Could not read from data directory: {exc}"
                ) from exc

        # NOTE we can skip this if we have a way to ship MLX
        # PyTorch safetensors to MLX safetensors
        model_dir_local = model_path.replace("/", "-")
        model_dir_local = f"{ckpt_output_dir}/{model_dir_local}"
        model_dir_mlx = f"{model_dir_local}-mlx"
        model_dir_mlx_quantized = f"{model_dir_local}-mlx-q"

        if skip_quantize:
            dest_model_dir = model_dir_mlx
            quantize_arg = False
        else:
            dest_model_dir = model_dir_mlx_quantized
            quantize_arg = True

        if tokenizer_dir is not None and gguf_model_path is not None:
            if not local:
                tokenizer_dir_local = tokenizer_dir.replace("/", "-")
                fetch_tokenizer_from_hub(tokenizer_dir, tokenizer_dir_local)

            # no need to pass quantize_arg for now, script automatically detects if quantization is necessary based on whether gguf model is quantized or not
            load(
                gguf=gguf_model_path,
                repo=tokenizer_dir,
                mlx_path=dest_model_dir,
            )

            for filename in os.listdir(model_dir_local):
                shutil.copy(
                    os.path.join(model_dir_local, filename),
                    os.path.join(dest_model_dir, filename),
                )
            shutil.rmtree(model_dir_local, ignore_errors=True)

        else:
            # Downloading PyTorch SafeTensor and Converting to MLX SafeTensor
            convert_between_mlx_and_pytorch(
                hf_path=model_path,
                mlx_path=dest_model_dir,
                quantize=quantize_arg,
                local=local,
            )

        adapter_file_path = f"{dest_model_dir}/adapters.npz"

        # train the model with LoRA
        load_and_train(
            model=dest_model_dir,
            train=True,
            data=data_path,
            adapter_file=adapter_file_path,
            iters=iters,
            save_every=10,
            steps_per_eval=10,
        )
    else:
        # Local
        from ..llamacpp.llamacpp_convert_to_gguf import convert_llama_to_gguf
        from ..train.linux_train import linux_train

        training_results_dir = linux_train(
            train_file=train_file,
            test_file=test_file,
            model_name=model_path,
            num_epochs=num_epochs,
            train_device=device,
            four_bit_quant=four_bit_quant,
        )

        final_results_dir = training_results_dir / "final"
        if final_results_dir.exists():
            shutil.rmtree(final_results_dir)
        final_results_dir.mkdir()

        gguf_models_dir = Path(DEFAULTS.CHECKPOINTS_DIR)
        gguf_models_dir.mkdir(exist_ok=True)
        gguf_models_file = gguf_models_dir / "ggml-model-f16.gguf"

        # Remove previously trained model, its taking up space we may need in the next step
        gguf_models_file.unlink(missing_ok=True)

        # TODO: Figure out what to do when there are multiple checkpoint dirs.
        # Right now it's just copying files from the first one numerically not necessarily the best one
        for fpath in (
            "checkpoint-*/added_tokens.json",
            "checkpoint-*/special_tokens_map.json",
            "checkpoint-*/tokenizer.json",
            "checkpoint-*/tokenizer.model",
            "checkpoint-*/tokenizer_config.json",
            "merged_model/config.json",
            "merged_model/generation_config.json",
        ):
            file_ = next(training_results_dir.glob(fpath))
            shutil.copy(file_, final_results_dir)
            logger.info(f"Copied {file_} to {final_results_dir}")

        for file in training_results_dir.glob("merged_model/*.safetensors"):
            shutil.move(file, final_results_dir)
            logger.info(f"Moved {file} to {final_results_dir}")

        if four_bit_quant:
            logger.info(
                "SKIPPING CONVERSION to gguf. This is unsupported with --4-bit-quant. "
                + "See https://github.com/instructlab/instructlab/issues/579."
            )
            return

        gguf_file_path = convert_llama_to_gguf(model=final_results_dir, pad_vocab=True)

        # Remove safetensors files to save space, were done with them here
        # and the huggingface lib has them cached
        for file in final_results_dir.glob("*.safetensors"):
            file.unlink()

        shutil.move(gguf_file_path, gguf_models_file)
        logger.info(f"Save trained model to {gguf_models_file}")



# Standard
from unittest import mock
from unittest.mock import MagicMock, patch
import os
import pathlib

# Third Party
from click.testing import CliRunner
import click
import pytest

# First Party
from instructlab.cli.model.serve import warn_for_unsupported_backend_param
from instructlab.model.backends.vllm import get_max_stable_vram_wait

# Local
from . import common


def assert_vllm_args(args):
    assert "python" in args[0]
    assert args[1] == "-m"
    assert args[2] == "vllm.entrypoints.openai.api_server"


def assert_template(args, expect_chat, path_chat, chat_value):
    hit_chat = False
    template_exists = False
    template = ""
    for s in args:
        if hit_chat:
            template = s
            template_exists = pathlib.Path(s).exists()
            break  # break as soon as we find the chat template otherwise we will get the next argument and template_exists will fail
        if s == "--chat-template":
            hit_chat = True

    assert hit_chat == expect_chat
    if path_chat:
        assert template_exists
        assert template == chat_value


def test_chat_auto(cli_runner: CliRunner):
    args = common.vllm_setup_test(
        cli_runner,
        ["--config=DEFAULT", "model", "serve", "--model-path=foo"],
    )
    assert_vllm_args(args)
    assert_template(args=args, expect_chat=True, path_chat=False, chat_value="")


def test_chat_custom(cli_runner: CliRunner, tmp_path: pathlib.Path):
    file = tmp_path / "chat_template.jinja2"
    file.touch()
    args = common.vllm_setup_test(
        cli_runner,
        [
            "--config=DEFAULT",
            "model",
            "serve",
            "--model-path=foo",
            "--chat-template={}".format(str(file)),
        ],
    )
    assert_vllm_args(args)
    assert_template(args=args, expect_chat=True, path_chat=True, chat_value=str(file))


def test_chat_manual(cli_runner: CliRunner):
    args = common.vllm_setup_test(
        cli_runner,
        [
            "--config=DEFAULT",
            "model",
            "serve",
            "--model-path=foo",
            "--chat-template=tokenizer",
        ],
    )
    assert_vllm_args(args)
    assert_template(args=args, expect_chat=False, path_chat=False, chat_value="")


def test_gpus(cli_runner: CliRunner):
    args = common.vllm_setup_test(
        cli_runner,
        [
            "--config=DEFAULT",
            "model",
            "serve",
            "--model-path=foo",
            "--gpus",
            "8",
        ],
    )
    common.assert_tps(args, "8")


def test_gpus_default(cli_runner: CliRunner):
    args = common.vllm_setup_test(
        cli_runner,
        [
            "--config=DEFAULT",
            "model",
            "serve",
            "--model-path=foo",
        ],
    )
    assert "--tensor-parallel-size" not in args


def test_ctx_tps_with_extra_params(cli_runner: CliRunner):
    args = common.vllm_setup_test(
        cli_runner,
        [
            "--config=DEFAULT",
            "model",
            "serve",
            "--model-path=foo",
            "--",
            "--served-model-name",
            "mymodel",
            "--tensor-parallel-size",
            "8",
        ],
    )
    assert args[-4] == "--served-model-name"
    assert args[-3] == "mymodel"
    common.assert_tps(args, "8")


def test_ctx_tps_with_gpus(cli_runner: CliRunner):
    args = common.vllm_setup_test(
        cli_runner,
        [
            "--config=DEFAULT",
            "model",
            "serve",
            "--model-path=foo",
            "--gpus",
            "8",
            "--",
            "--served-model-name",
            "mymodel",
            "--tensor-parallel-size",
            "4",
        ],
    )
    assert args[-4] == "--served-model-name"
    assert args[-3] == "mymodel"
    common.assert_tps(args, "4")


def test_gpus_config(cli_runner: CliRunner):
    fname = common.setup_gpus_config(gpus=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
        ],
    )
    common.assert_tps(args, "8")


def test_gpus_with_gpus_config(cli_runner: CliRunner):
    fname = common.setup_gpus_config(gpus=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
            "--gpus",
            "4",
        ],
    )
    common.assert_tps(args, "4")


def test_ctx_tps_with_gpus_config(cli_runner: CliRunner):
    fname = common.setup_gpus_config(gpus=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
            "--",
            "--tensor-parallel-size",
            "4",
        ],
    )
    common.assert_tps(args, "4")


def test_gpus_with_ctx_tps_with_gpus_config(cli_runner: CliRunner):
    fname = common.setup_gpus_config(gpus=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
            "--gpus",
            "2",
            "--",
            "--tensor-parallel-size",
            "4",
        ],
    )
    common.assert_tps(args, "4")


def test_vllm_args_config(cli_runner: CliRunner):
    fname = common.setup_gpus_config(tps=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
        ],
    )
    common.assert_tps(args, "8")


def test_vllm_args_config_with_gpus_config(cli_runner: CliRunner):
    fname = common.setup_gpus_config(gpus=4, tps=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
        ],
    )
    common.assert_tps(args, "4")


def test_vllm_args_config_with_gpus(cli_runner: CliRunner):
    fname = common.setup_gpus_config(tps=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
            "--gpus",
            "4",
        ],
    )
    common.assert_tps(args, "4")


def test_vllm_args_config_with_ctx_tps(cli_runner: CliRunner):
    fname = common.setup_gpus_config(tps=8)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
            "--",
            "--tensor-parallel-size",
            "4",
        ],
    )
    common.assert_tps(args, "4")


def test_vllm_args_null(cli_runner: CliRunner):
    fname = common.setup_gpus_config(vllm_args=lambda: None)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "serve",
            "--model-path=foo",
            "--gpus",
            "4",
        ],
    )
    common.assert_tps(args, "4")


def test_max_stable_vram_wait():
    with mock.patch.dict(os.environ, {"ILAB_MAX_STABLE_VRAM_WAIT": "0"}):
        wait = get_max_stable_vram_wait(10)
        assert wait == 0
    with mock.patch.dict(os.environ, clear=True):
        wait = get_max_stable_vram_wait(10)
        assert wait == 10


@pytest.mark.parametrize(
    "param,expected_call_count",
    [
        ("gpu_layers", 1),
        ("num_threads", 1),
        ("max_ctx_size", 1),
        (
            "supported_param",
            0,
        ),  # Example of a parameter that should not trigger a warning
    ],
)
def test_warn_for_unsupported_backend_param(param, expected_call_count):
    with patch("instructlab.cli.model.serve.logger.warning") as mock_warning:
        # Create a mock click.Context object
        ctx = MagicMock()

        # Set the get_parameter_source to return COMMANDLINE for the tested param
        ctx.get_parameter_source.side_effect = (
            lambda x: click.core.ParameterSource.COMMANDLINE if x == param else None
        )

        # Call the function to test
        warn_for_unsupported_backend_param(ctx)

        # Assert that logger.warning was called the expected number of times
        assert mock_warning.call_count == expected_call_count
        if expected_call_count > 0:
            mock_warning.assert_called_with(
                f"Option '--{param.replace('_','-')}' not supported by the backend."
            )



#!/usr/bin/env python3
# SPDX-License-Identifier: Apache-2.0

"""
This script validates a Jupyter notebook file.
Usage: python validate_notebook.py <notebook.ipynb>
"""

# Standard
import sys

# Third Party
import nbformat


def validate_notebook(file_path):
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            nb = nbformat.read(f, as_version=4)
        nbformat.validate(nb)
        print(f"{file_path} is a valid Jupyter notebook.")
        return True
    except nbformat.ValidationError as e:
        print(f"{file_path} is not a valid Jupyter notebook. Validation error: {e}")
        return False
    except Exception as e:
        print(f"An error occurred while validating {file_path}: {e}")
        return False


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python validate_notebook.py <notebook.ipynb>")
        sys.exit(1)

    notebook_path = sys.argv[1]
    if not validate_notebook(notebook_path):
        sys.exit(1)



# SPDX-License-Identifier: Apache-2.0
"""Verify package metadata"""

# Standard
from importlib.metadata import metadata
import typing

# Third Party
from packaging.requirements import Requirement
from packaging.version import Version
import pytest

PKG_NAME = "instructlab"
# expected hardware extras
HW_EXTRAS = frozenset({"cpu", "cuda", "hpu", "mps", "rocm"})
# special cases
EXTRA_CHECKS = {
    "hpu": {
        "torch": Version("2.3.1a0"),
        "transformers": Version("4.43.0"),
    }
}


def iter_requirements(pkg: str = PKG_NAME) -> typing.Iterable[Requirement]:
    m = metadata(pkg)
    requires = m.get_all("Requires-Dist")
    assert requires is not None
    for r in requires:
        yield Requirement(r)


def test_provides_extra():
    m = metadata(PKG_NAME)
    assert set(m.get_all("Provides-Extra")).issuperset(HW_EXTRAS)


def test_require_no_url_req():
    # PyPI does not accept packages with URL requirements
    for req in iter_requirements():
        assert req.url is None, req


@pytest.mark.parametrize("hw_extra", sorted(HW_EXTRAS))
@pytest.mark.parametrize("py_version", ["3.10", "3.11"])
def test_package_conflict(py_version: str, hw_extra: str) -> None:
    if py_version != "3.11" and hw_extra == "hpu":
        pytest.skip("Intel Gaudi only supports 3.11")

    base: dict[str, Requirement] = {}
    hw: dict[str, Requirement] = {}
    for req in iter_requirements():
        # override version for environment
        base_env = {
            "implementation_version": f"{py_version}.0",
            "python_full_version": f"{py_version}.0",
            "python_version": py_version,
        }
        extra_env = base_env.copy()
        extra_env["extra"] = hw_extra

        if req.marker is None or req.marker.evaluate(base_env):
            # no marker or no optional requirement
            base[req.name] = req
        elif req.marker.evaluate(extra_env):
            # matching optional requirement
            hw[req.name] = req

    for name, hwreq in hw.items():
        basereq = base.get(name)
        if basereq is None:
            continue
        for specifier in hwreq.specifier:
            # naive check for common version conflicts
            # allow pre-releases for Gaudi
            if specifier.operator in {"~=", "==", "<=", ">="}:
                version: Version = Version(specifier.version)
                assert basereq.specifier.contains(
                    version, prereleases=version.is_prerelease
                ), (basereq, hwreq)

    # verify special cases against base requirements
    if hw_extra in EXTRA_CHECKS:
        for name, basereq in base.items():
            extra_check = EXTRA_CHECKS[hw_extra].get(name)
            if extra_check is not None:
                assert basereq.specifier.contains(
                    extra_check, prereleases=extra_check.is_prerelease
                ), (basereq, extra_check)



# SPDX-License-Identifier: Apache-2.0

# Standard
from unittest.mock import Mock, patch
import typing

# Third Party
import git
import pytest
import yaml

# First Party
from instructlab import utils
from instructlab.utils import Message, MessageSample


class TestUtils:
    """Test collection in instructlab.utils."""

    @patch(
        "instructlab.utils.git_clone_checkout",
        return_value=Mock(
            spec=git.Repo, working_dir="tests/testdata/temp_taxonomy_repo"
        ),
    )
    def test_validate_documents(self, git_clone_checkout):
        with open(
            "tests/testdata/knowledge_valid.yaml", "r", encoding="utf-8"
        ) as qnafile:
            utils._validate_documents(
                source=yaml.safe_load(qnafile).get("document"),
                skip_checkout=True,
            )
            git_clone_checkout.assert_called_once()

    def test_convert_to_legacy_from_pretraining_messages(
        self,
    ):
        new_dataset: typing.List[MessageSample] = [
            {
                "messages": [
                    {
                        "role": "system",
                        "content": "You are a friendly assistant",
                    },
                    {
                        "role": "pretraining",
                        "content": "<|user|>What is 2+2?<|assistant|>2+2=4",
                    },
                ],
                "group": " test",
                "dataset": "test-dataset",
                "metadata": "test",
            }
        ]
        legacy = utils.ensure_legacy_dataset(new_dataset)
        assert len(legacy) == 1
        assert legacy[0]["system"] == "You are a friendly assistant"
        assert legacy[0]["user"] == "What is 2+2?"
        assert legacy[0]["assistant"] == "2+2=4"

    @pytest.mark.parametrize(
        "content,exception,match",
        [
            ("<|user|>What is 2+2? 2+2=4", ValueError, "<|assistant|>"),
            ("<|assistant|>2+2=4", ValueError, "<|user|>"),
            ("<|user|>what is 2+2?<|assistant|>2+2=4", None, ""),
        ],
    )
    def test_invalid_pretraining_messages(
        self, content: str, exception: Exception | None, match: str
    ):
        new_dataset: typing.List[MessageSample] = [
            {
                "messages": [
                    {
                        "role": "system",
                        "content": "You are a friendly assistant",
                    },
                    {
                        "role": "pretraining",
                        "content": content,
                    },
                ],
                "group": " test",
                "dataset": "test-dataset",
                "metadata": "test",
            }
        ]
        if exception:
            with pytest.raises(ValueError, match=match):
                utils.ensure_legacy_dataset(new_dataset)
        else:
            utils.ensure_legacy_dataset(new_dataset)

    def test_pretraining_messages_without_system(self):
        new_dataset: typing.List[MessageSample] = [
            {
                "messages": [
                    {
                        "role": "pretraining",
                        "content": "<|user|>What is 2+2?<|assistant|>2+2=4",
                    },
                ],
                "group": " test",
                "dataset": "test-dataset",
                "metadata": "test",
            }
        ]
        legacy = utils.ensure_legacy_dataset(new_dataset)
        assert len(legacy) == 1
        assert legacy[0]["system"] == ""

    def test_convert_to_legacy_from_messages(self):
        messages: typing.List[MessageSample] = [
            {
                "messages": [
                    {
                        "role": "system",
                        "content": "You are a friendly assistant",
                    },
                    {"role": "user", "content": "Who is pickle rick?"},
                    {
                        "role": "assistant",
                        "content": "As an AI language model, I have absolutely no idea.",
                    },
                ],
                "group": " test",
                "dataset": "test-dataset",
                "metadata": "test",
            }
        ]
        legacy = utils.ensure_legacy_dataset(messages)
        assert len(legacy) == 1
        sample = legacy[0]
        assert sample["system"] == "You are a friendly assistant"
        assert sample["user"] == "Who is pickle rick?"
        assert (
            sample["assistant"] == "As an AI language model, I have absolutely no idea."
        )

    @pytest.mark.parametrize(
        "system,user,assistant",
        [
            (None, None, None),
            ("You are a friendly assistant trained by ACME corp", None, None),
            (None, "Who is pickle rick?", None),
            (
                "You are a friendly assistant trained by ACME corp",
                "Who is pickle rick?",
                None,
            ),
            (None, None, "As an AI language model, I have absolutely no idea."),
            (
                "You are a friendly assistant trained by ACME corp",
                None,
                "As an AI language model, I have absolutely no idea.",
            ),
            (
                None,
                "Who is pickle rick?",
                "As an AI language model, I have absolutely no idea.",
            ),
        ],
    )
    def test_invalid_datasets(
        self, system: str | None, user: str | None, assistant: str | None
    ):
        messages: typing.List[Message] = []
        if system:
            messages.append({"content": system, "role": "system"})
        if user:
            messages.append({"content": user, "role": "user"})
        if assistant:
            messages.append({"content": assistant, "role": "assistant"})
        dataset: typing.List[MessageSample] = [
            {
                "messages": messages,
                "group": "ACME",
                "dataset": "The Pickle Rick Collection",
                "metadata": "{{ pickle: rick, }}",
            }
        ]
        with pytest.raises(ValueError):
            utils.ensure_legacy_dataset(dataset)


@pytest.mark.parametrize(
    "url,expected_host,expected_port",
    [
        ("127.0.0.1:8080", "127.0.0.1", 8080),
        ("[::1]:8080", "::1", 8080),
        ("host.test:9090", "host.test", 9090),
        ("https://host.test:443/egg/spam", "host.test", 443),
    ],
)
def test_split_hostport(url, expected_host, expected_port):
    host, port = utils.split_hostport(url)
    assert host == expected_host
    assert port == expected_port


@pytest.mark.parametrize(
    "url",
    [
        "127.0.0.1",
        "",
        "::1:8080",
    ],
)
def test_split_hostport_err(url):
    with pytest.raises(ValueError):
        utils.split_hostport(url)



# Standard
from dataclasses import dataclass
import abc
import logging
import pathlib
import typing

# Third Party
import httpx

# Local
from ...log import add_file_handler_to_logger
from .common import CHAT_TEMPLATE_AUTO, Closeable, safe_close_all

logger = logging.getLogger(__name__)


@dataclass
class ServerConfig:
    api_base: str
    log_file: typing.Optional[pathlib.Path] = None


class BackendServer(abc.ABC):
    """Base class for a serving backend"""

    def __init__(
        self,
        model_family: str,
        model_path: pathlib.Path,
        chat_template: str,
        host: str,
        port: int,
        config: ServerConfig,
    ) -> None:
        self.model_family = model_family
        self.model_path = model_path
        self.chat_template = chat_template if chat_template else CHAT_TEMPLATE_AUTO
        self.host = host
        self.port = port
        self.config = config
        self.resources: list[Closeable] = []

        # Write logs to a file if log_file is provided
        if self.config.log_file:
            self.setup_logs_to_file(logger)

    def setup_logs_to_file(self, a_logger: logging.Logger) -> None:
        """Write logs to a file if log_file is provided"""
        add_file_handler_to_logger(a_logger, self.config.log_file)

    @abc.abstractmethod
    def run(self):
        """Run serving backend in foreground (ilab model serve)"""

    @abc.abstractmethod
    def run_detached(
        self,
        http_client: httpx.Client | None = None,
        background: bool = True,
        foreground_allowed: bool = False,
        max_startup_retries: int = 0,
    ) -> str:
        """Run serving backend in background ('ilab model chat' when server is not running)"""

    def shutdown(self):
        """Shutdown serving backend"""

        safe_close_all(self.resources)

    def register_resources(self, resources: typing.Iterable[Closeable]) -> None:
        self.resources.extend(resources)

    @abc.abstractmethod
    def get_backend_type(self):
        """Return which type of backend this is, llama-cpp or vllm"""



# Standard
import logging

# Third Party
from instructlab.sdg.generate_data import generate_data

# pylint: disable=ungrouped-imports
from instructlab.sdg.utils import GenerateException
import openai

# First Party
from instructlab.utils import HttpClientParams, http_client

logger = logging.getLogger(__name__)


def gen_data(
    serve_cfg,
    model_path,
    num_cpus,
    sdg_scale_factor,
    taxonomy_path,
    taxonomy_base,
    output_dir,
    quiet,
    endpoint_url,
    api_key,
    yaml_rules,
    chunk_word_count,
    server_ctx_size,
    http_client_params: HttpClientParams,
    model_family,
    pipeline,
    enable_serving_output,
    batch_size,
    gpus,
    checkpoint_dir,
):
    """Generates synthetic data to enhance your example data"""
    backend_instance = None

    if endpoint_url:
        api_base = endpoint_url
    else:
        # First Party
        from instructlab.model.backends import backends
        from instructlab.model.backends.llama_cpp import Server as llama_cpp_server

        backend_instance = backends.select_backend(cfg=serve_cfg, model_path=model_path)
        if (
            backend_instance.get_backend_type() is not backends.VLLM
            and gpus is not None
        ):
            logger.debug(
                "Cannot specify '--gpus' with a llama-cpp backend, ignoring this flag."
            )

        try:
            # Run the backend server
            api_base = backend_instance.run_detached(
                http_client=http_client(http_client_params),
                background=not enable_serving_output,
                foreground_allowed=True,
                max_startup_retries=1,
            )
        except Exception as exc:
            raise ValueError(f"Failed to start server: {exc}") from exc

        # disable batching when running with the local llama.cpp server
        if isinstance(backend_instance, llama_cpp_server):
            if batch_size is not None:
                logger.warning(
                    "Disabling SDG batching - unsupported with llama.cpp serving"
                )
            batch_size = 0

    client = openai.OpenAI(
        base_url=api_base, api_key=api_key, http_client=http_client(http_client_params)
    )

    try:
        logger.info(
            f"Generating synthetic data using '{pipeline}' pipeline, '{model_path}' model, '{taxonomy_path}' taxonomy, against {api_base} server"
        )
        generate_data(
            client=client,
            model_family=model_family,
            model_name=model_path,
            num_cpus=num_cpus,
            num_instructions_to_generate=sdg_scale_factor,
            taxonomy=taxonomy_path,
            taxonomy_base=taxonomy_base,
            output_dir=output_dir,
            console_output=not quiet,
            yaml_rules=yaml_rules,
            chunk_word_count=chunk_word_count,
            server_ctx_size=server_ctx_size,
            pipeline=pipeline,
            batch_size=batch_size,
            checkpoint_dir=checkpoint_dir,
        )
    except GenerateException as exc:
        raise ValueError(
            f"Generating dataset failed with the following error: {exc}"
        ) from exc
    finally:
        if backend_instance is not None:
            backend_instance.shutdown()



# SPDX-License-Identifier: Apache-2.0
"""Click extensions for InstructLab"""

# Standard
from importlib import metadata
import enum
import functools
import json
import logging
import os
import sys
import typing

# Third Party
from click.core import ParameterSource
from click_didyoumean import DYMGroup
import click

# First Party
from instructlab.configuration import DEFAULTS, BaseModel, storage_dirs_exist

logger = logging.getLogger(__name__)


class LazyEntryPointGroup(DYMGroup):
    """Lazy load commands from an entry point group

    Entry points are defined in `pyproject.toml`. Example:

        [project.entry-points."instructlab.command.config"]
        "init" = "instructlab.config.init:init"

    This defines the command `ilab config init` with the function
    `from instructlab.config.init import init` as click command.
    """

    def __init__(self, *args, ep_group: str, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.eps: metadata.EntryPoints = metadata.entry_points(group=ep_group)
        if not self.eps.names:
            raise ValueError(f"{ep_group} is empty")

    def list_commands(self, ctx: click.Context) -> list[str]:
        result = list(super().list_commands(ctx))
        result.extend(sorted(self.eps.names))
        return result

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if cmd_name in self.eps.names:
            cmd = self.eps[cmd_name].load()
        else:
            cmd = super().get_command(ctx, cmd_name)
        if typing.TYPE_CHECKING:
            assert isinstance(cmd, click.Command) or cmd is None
        return cmd


class ExpandAliasesGroup(LazyEntryPointGroup):
    """Lazy load commands and aliases from entry point groups"""

    def __init__(self, *args, alias_ep_group: str, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.alias_eps: metadata.EntryPoints = metadata.entry_points(
            group=alias_ep_group
        )

    def get_alias_info(self, cmd_name: str) -> tuple[str, str]:
        ep: metadata.EntryPoint = self.alias_eps[cmd_name]
        # assume that the second item of the module name is the group
        return ep.module.split(".", 3)[1], ep.module.split(".", 3)[2]

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if cmd_name in self.alias_eps.names:
            cmd = self.alias_eps[cmd_name].load()
            if typing.TYPE_CHECKING:
                assert isinstance(cmd, click.Command)
            # if some storage dirs do not exist
            # AND the --config flag is not customized, then we error
            if not storage_dirs_exist() and (
                ctx.params["config_file"] == DEFAULTS.CONFIG_FILE
            ):
                click.secho(
                    "Some ilab storage directories do not exist yet. Please run `ilab config init` before continuing.",
                    fg="red",
                )
                raise click.exceptions.Exit(1)
            return cmd
        return super().get_command(ctx, cmd_name)

    def format_epilog(self, ctx: click.Context, formatter):
        """Inject our aliases into the help string"""
        if self.alias_eps.names:
            formatter.write_paragraph()
            formatter.write_text("Aliases:")
            with formatter.indentation():
                cmd_names = sorted(self.alias_eps.names)
                pad = max(len(cmd_name) for cmd_name in cmd_names)
                for cmd_name in cmd_names:
                    group, primary = self.get_alias_info(cmd_name)
                    formatter.write_text(f"{cmd_name:<{pad}}  {group} {primary}\n")

        super().format_epilog(ctx, formatter)


class ConfigOption(click.Option):
    """A click.Option with extended default lookup help

    The defaults and description can be looked up in a subsection of the context's
    config object.

    The help output of ConfigOption class reads defaults from the context's
    config object field and shows the name of the config settings.

    Example:
        Options:
        --model-path PATH           Path to the model used during generation.
                                    [default: models/merlinite-7b-lab-Q4_K_M.gguf;
                                    config: 'serve.model_path']
    """

    def __init__(
        self,
        *args: typing.Any,
        config_sections: str | None = None,
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        if self.show_default:
            raise ValueError("show_default must be False")
        if self.help:
            raise ValueError(
                f"help must not be set for '{self.name}', it is derived from the Field description in the Config class"
            )
        if config_sections:
            self.config_sections = list(config_sections.split("."))
        else:
            self.config_sections = []

    def get_help_record(self, ctx: click.Context) -> tuple[str, str] | None:
        result = super().get_help_record(ctx)
        if result is None or ctx.default_map is None:
            # hidden is True
            # missing default map (clickman, sphinx)
            return result

        cmd = ctx.command.name
        name = self.name

        if self.config_sections:
            config_identifier = (
                [str(cmd)]
                + [
                    str(section)
                    for section in self.config_sections
                    if section is not None
                ]
                + [str(name)]
            )
        else:
            config_identifier = [str(cmd), str(name)]

        if typing.TYPE_CHECKING:
            assert name is not None

        # create help extra
        description, default_value = get_default_and_description(
            ctx.obj.config, config_identifier
        )
        if default_value is None:
            default_string = "<None>"
        elif isinstance(default_value, (list, tuple)):
            default_string = ", ".join(str(d) for d in default_value)
        elif isinstance(default_value, enum.Enum):
            default_string = default_value.value
        else:
            default_string = str(default_value)

        default_msg = (
            f"default: {default_string}; config: '{'.'.join(config_identifier)}'"
        )

        # extend message
        prefix, msg = result
        # If the 'help' field on click.Option is not set, then 'msg' will be None
        # unless options like 'required' are set
        if msg:
            # Needed to handle [required] option for example
            if msg.endswith("]"):
                msg = f"{description} {msg[:-1]}; {default_msg}]"
            else:
                msg = f"{description} {msg} [{default_msg}]"
        else:
            # do not append description if msg is None to avoid space
            msg = f"{description} [{default_msg}]"  # type: ignore
        return prefix, msg

    def consume_value(
        self, ctx: click.Context, opts: typing.Mapping[str, typing.Any]
    ) -> tuple[typing.Any, ParameterSource]:
        value, source = super().consume_value(ctx, opts)
        # fix parameter source for config section that are mis-reported
        # as DEFAULT source instead of DEFAULT_MAP source.
        if (
            source == ParameterSource.DEFAULT
            and self.config_sections
            and ctx.default_map is not None
            and self.name is not None
        ):
            section = ctx.default_map
            for secname in self.config_sections:
                section = section.get(secname, {})
            if self.name in section:
                source = ParameterSource.DEFAULT_MAP
        return value, source

    def get_default(
        self, ctx: click.Context, call: bool = True
    ) -> typing.Any | typing.Callable[[], typing.Any] | None:
        """Lookup default in config subsection

        Used so "serve" option "gpu_layers" looks up its default in
        "serve.llama_cpp.gpu_layers" instead of "serve.gpu_layers".
        """
        if not self.config_sections:
            # no config subsection
            return super().get_default(ctx, call=call)
        if ctx.default_map is not None and self.name is not None:
            section = ctx.default_map
            for secname in self.config_sections:
                section = section.get(secname, {})
            value = section.get(self.name)
            if call and callable(value):
                return value()
            return value
        return None


def _get_param_info(
    ctx: click.Context, **kwargs: dict
) -> typing.Generator[tuple[str, typing.Any, str, str], None, None]:
    """Get click parameter information

    Returns name, value, type name, and parameter source
    """
    for key, value in kwargs.items():
        src = ctx.get_parameter_source(key)
        param_src: str = src.name.lower() if src is not None else "unknown"
        type_name: str
        if value is None:
            type_name = "None"
        else:
            type_: type = type(value)
            type_name = type_.__name__
            mod_name: str | None = getattr(type_, "__module__", None)
            if mod_name and mod_name != "builtins":
                type_name = f"{mod_name}.{type_name}"
        yield key, value, type_name, param_src
    # additional args
    if ctx.args:
        yield "args", ctx.args, "list", "args"


class _ParamEncoder(json.JSONEncoder):
    """Custom encoder for additional parameter types"""

    def default(self, o: typing.Any) -> typing.Any:
        if isinstance(o, os.PathLike):
            return os.fsdecode(o)
        return super().default(o)


def display_params(f: typing.Callable) -> typing.Callable:
    """Display command parameters decorator

    The function decorator adds two hidden parameters to a click Command.

    - `--debug-params` dumps parameter info for humans and exits immediately.
    - `--display-param-json` dumps parameter info as JSON and exits immediately.

    When DEBUG logging is enabled, it also dumps parameter information for
    humans, but does not exit immediately. The output contains parameter
    name, value, value type (as qualified dotted string), and the source of
    the value (commandline, environment, default, default_map, args,
    unknown).
    """

    @functools.wraps(f)
    def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        # some callbacks don't need ctx
        ctx: click.Context | None = kwargs.get("ctx")
        if ctx is None:
            ctx = click.get_current_context()
        display: str = kwargs.pop("_debug_params")

        # also display params when debug logging is enabled
        if (
            display is None
            and logger.isEnabledFor(logging.DEBUG)
            and not kwargs.get("quiet", False)
        ):
            display = "logger"

        if kwargs and display in {"logger", "human"}:
            pad = min(max(len(name) for name in kwargs), 24)
            print("Parameters:")
            for name, value, typ, src in _get_param_info(ctx, **kwargs):
                print(f"  {name:>{pad}}: {value!r:<8} \t[type: {typ}, src: {src}]")
        elif display == "json":
            params = {
                name: (value, type_name, param_src)
                for name, value, type_name, param_src in _get_param_info(ctx, **kwargs)
            }
            json.dump({"params": params}, sys.stdout, cls=_ParamEncoder)

        # --debug-params* exit
        if display in {"json", "human"}:
            ctx.exit()

        return f(*args, **kwargs)

    human_option = click.option(
        "--debug-params",
        "_debug_params",
        flag_value="human",
        hidden=True,
        help="display parameter info and exit.",
    )
    json_option = click.option(
        "--debug-params-json",
        "_debug_params",
        flag_value="json",
        hidden=True,
        help="display parameter as JSON and exit.",
    )

    return human_option(json_option(wrapper))


def get_default_and_description(
    cfg: BaseModel, config_identifier: list[str]
) -> typing.Tuple[str | None, typing.Any]:
    """
    Retrieve the default value and description for a given configuration field name.

    This function searches through the fields of a Pydantic BaseModel to find a field
    that matches the provided configuration name. If the field is found, it returns
    the field's description and default value. If the field is a nested model, the
    function is called recursively to search within the nested model.

    Args:
        cfg (BaseModel): The Pydantic model instance containing the configuration fields.
        config_identifier (list[str]): A list of field names to search for in the model.

    Returns:
        typing.Tuple[str | None, typing.Any]: A tuple containing the field's description
        and default value. If the field is not found, a ValueError is raised.

    Raises:
        ValueError: If the specified config_identifier is not found in the model.
    """
    # Loop through the fields of the model
    for field_name, field in cfg.model_fields.items():
        if field_name == config_identifier[0]:
            value = getattr(cfg, field_name)
            description = field.description
            default_value = field.get_default(call_default_factory=True)

            # If the value is a nested model and there are more names to check, recurse
            # Slice the config_identifier list to remove the current field name
            if isinstance(value, BaseModel) and len(config_identifier) > 1:
                return get_default_and_description(value, config_identifier[1:])

            return description, default_value

    # If no match is found, raise an exception
    raise ValueError(f"{config_identifier} not in Config object")



# SPDX-License-Identifier: Apache-2.0

# Standard
import copy
import logging
import os.path

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab.configuration import DEFAULTS
from instructlab.data.generate_data import gen_data  # type: ignore
from instructlab.utils import HttpClientParams, contains_argument

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--model",
    "model_path",
    cls=clickext.ConfigOption,
    config_sections="teacher",
    show_default=False,
)
@click.option(
    "--num-cpus",
    type=click.INT,
    help="Number of processes to use.",
    default=DEFAULTS.NUM_CPUS,
    show_default=True,
)
@click.option(
    "--chunk-word-count",
    type=click.INT,
    help="Number of words to chunk the document",
    default=DEFAULTS.CHUNK_WORD_COUNT,
    show_default=True,
)
# TODO - DEPRECATED - Remove in a future release
@click.option(
    "--num-instructions",
    type=click.INT,
    default=-1,
    hidden=True,
)
@click.option(
    "--sdg-scale-factor",
    type=click.INT,
    help="Number of instructions to generate for each seed example. The examples map to sample q&a pairs for new skills. For knowledge, examples are generated with both the sample q&a pairs, as well as chunks of the knowledge document(s), so the resulting data set is typically larger for a knowledge addition for the same value of `--sdg-scale-factor`.",
    show_default=True,
)
@click.option(
    "--taxonomy-path",
    type=click.Path(),
    cls=clickext.ConfigOption,
)
@click.option(
    "--taxonomy-base",
    cls=clickext.ConfigOption,
)
@click.option(
    "--output-dir",
    type=click.Path(),
    cls=clickext.ConfigOption,
)
@click.option(
    "--quiet",
    is_flag=True,
    help="Suppress output of synthesized instructions.",
)
@click.option(
    "--endpoint-url",
    type=click.STRING,
    help="Custom URL endpoint for OpenAI-compatible API. Defaults to the `ilab model serve` endpoint.",
)
@click.option(
    "--api-key",
    type=click.STRING,
    default=DEFAULTS.API_KEY,  # Note: do not expose default API key
    help="API key for API endpoint. [default: config.DEFAULT_API_KEY]",
)
@click.option(
    "--yaml-rules",
    type=click.Path(),
    default=None,
    help="Custom rules file for YAML linting.",
)
@click.option(
    "--server-ctx-size",
    type=click.INT,
    default=DEFAULTS.MAX_CONTEXT_SIZE,
    show_default=True,
    help="The context size is the maximum number of tokens the server will consider.",
)
@click.option(
    "--tls-insecure",
    is_flag=True,
    help="Disable TLS verification.",
)
@click.option(
    "--tls-client-cert",
    type=click.Path(),
    default="",
    show_default=True,
    help="Path to the TLS client certificate to use.",
)
@click.option(
    "--tls-client-key",
    type=click.Path(),
    default="",
    show_default=True,
    help="Path to the TLS client key to use.",
)
@click.option(
    "--tls-client-passwd",
    type=click.STRING,
    default="",
    help="TLS client certificate password.",
)
@click.option(
    "--model-family",
    type=click.STRING,
    help="Force model family to use when picking a generation template",
)
@click.option(
    "--pipeline",
    type=click.STRING,
    cls=clickext.ConfigOption,
)
@click.option(
    "--batch-size",
    type=click.IntRange(min=0),
    default=None,
    help="Number of elements to process in each batch through the SDG pipeline. Enabled by default for the vLLM serving backend, with a batch size of 8 chosen based on experiments to optimize for throughput. Use 0 to disable.",
)
@click.option(
    "--enable-serving-output",
    is_flag=True,
    help="Print serving engine logs.",
)
@click.option(
    "--gpus",
    type=click.IntRange(min=0),
    cls=clickext.ConfigOption,
    config_sections="teacher.vllm",
)
@click.pass_context
@clickext.display_params
def generate(
    ctx,
    model_path,
    num_cpus,
    num_instructions,
    sdg_scale_factor,
    taxonomy_path,
    taxonomy_base,
    output_dir,
    quiet,
    endpoint_url,
    api_key,
    yaml_rules,
    chunk_word_count,
    server_ctx_size,
    tls_insecure: bool,
    tls_client_cert: str | None,
    tls_client_key: str | None,
    tls_client_passwd: str | None,
    model_family,
    pipeline,
    enable_serving_output,
    batch_size,
    gpus,
):
    """Generates synthetic data to enhance your example data"""

    # if --pipeline is not used, pipeline defaults to the value of ctx.obj.config.generate.pipeline
    # set in the config file. A user could intentionally set this to 'null' in the config file
    # if they want to ensure --pipeline needs to be used.
    # This would happen if the type of pipeline needs to be different across different runs of
    # `ilab data generate`.
    if not pipeline:
        click.secho(
            "Pipeline not set. Please use the --pipeline flag or set it in the config file.",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    if num_instructions != -1:
        click.secho(
            "The --num-instructions flag is deprecated. Please use --sdg-scale-factor instead.",
            fg="yellow",
        )

    # If batch size is not set explicitly, default to 8
    # Once https://github.com/instructlab/sdg/issues/224 is resolved we can
    # pass batch_size=None to the library instead
    if batch_size is None:
        batch_size = 8

    # Specify checkpoint dir if batching is enabled
    checkpoint_dir = None
    if batch_size > 0:
        checkpoint_dir = os.path.join(output_dir, "checkpoints")

    serve_cfg = copy.deepcopy(ctx.obj.config.generate.teacher)
    serve_cfg.llama_cpp.llm_family = model_family or serve_cfg.llama_cpp.llm_family
    serve_cfg.vllm.llm_family = model_family or serve_cfg.vllm.llm_family
    serve_cfg.vllm.vllm_args = serve_cfg.vllm.vllm_args or []
    if gpus is not None:
        tps_prefix = "--tensor-parallel-size"
        if contains_argument(tps_prefix, serve_cfg.vllm.vllm_args):
            click.secho(
                "Using gpus from --gpus. Ignoring --tensor-parallel-size configured in generate.teacher vllm_args",
                fg="yellow",
            )
        serve_cfg.vllm.vllm_args.extend([tps_prefix, str(gpus)])

    http_client_params = HttpClientParams(
        {
            "tls_client_cert": tls_client_cert,
            "tls_client_key": tls_client_key,
            "tls_client_passwd": tls_client_passwd,
            "tls_insecure": tls_insecure,
        }
    )

    try:
        gen_data(
            serve_cfg,
            model_path,
            num_cpus,
            sdg_scale_factor,
            taxonomy_path,
            taxonomy_base,
            output_dir,
            quiet,
            endpoint_url,
            api_key,
            yaml_rules,
            chunk_word_count,
            server_ctx_size,
            http_client_params,
            model_family,
            pipeline,
            enable_serving_output,
            batch_size,
            gpus,
            checkpoint_dir,
        )
    except Exception as exc:
        click.secho(f"failed to generate data with exception: {exc}", fg="red")
        raise click.exceptions.Exit(1)



# SPDX-License-Identifier: Apache-2.0

# Standard
from unittest.mock import MagicMock, patch

# Third Party
from click.testing import CliRunner
from huggingface_hub.utils import HfHubHTTPError

# First Party
from instructlab import lab


class TestLabDownload:
    # When using `from X import Y` you need to understand that Y becomes part
    # of your module, so you should use `my_module.Y`` to patch.
    # When using `import X`, you should use `X.Y` to patch.
    # https://docs.python.org/3/library/unittest.mock.html#where-to-patch?
    @patch("instructlab.model.download.hf_hub_download")
    def test_download(self, mock_hf_hub_download, cli_runner: CliRunner):
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "download",
                "--hf-token=foo",
            ],
        )
        assert (
            result.exit_code == 0
        ), f"command finished with an unexpected exit code. {result.stdout}"
        assert mock_hf_hub_download.call_count == 2

    @patch(
        "instructlab.model.download.hf_hub_download",
        MagicMock(side_effect=HfHubHTTPError("Could not reach hugging face server")),
    )
    def test_download_error(self, cli_runner: CliRunner):
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "download",
            ],
        )
        assert result.exit_code == 1, "command finished with an unexpected exit code"
        assert "Could not reach hugging face server" in result.output

    @patch("instructlab.model.download.OCIDownloader.download")
    def test_oci_download(self, mock_oci_download, cli_runner: CliRunner):
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "download",
                "--repository=docker://quay.io/ai-lab/models/granite-7b-lab",
                "--release=latest",
            ],
        )
        assert result.exit_code == 0
        mock_oci_download.assert_called_once()

    def test_oci_download_repository_error(self, cli_runner: CliRunner):
        result = cli_runner.invoke(
            lab.ilab,
            [
                "--config=DEFAULT",
                "model",
                "download",
                "--repository=docker://quay.io/ai-lab/models/granite-7b-lab:latest",
            ],
        )
        assert result.exit_code == 1
        assert (
            "Invalid repository supplied: Please specify tag/version 'latest' via --release"
            in result.output
        )
